"""SDMX XML StructureSpecificData reader aux module."""

import itertools
from typing import Any, Dict

import numpy as np
import pandas as pd

from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.__data_aux import (
    __process_df,
)
from pysdmx.io.xml.__tokens import (
    EXCLUDED_ATTRIBUTES,
    GROUP,
    OBS,
    SERIES,
)
from pysdmx.io.xml.utils import add_list
from pysdmx.model.dataset import ActionType


def _reading_str_series(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Series
    test_list = []
    df = None
    dataset[SERIES] = add_list(dataset[SERIES])
    for data in dataset[SERIES]:
        keys = dict(itertools.islice(data.items(), len(data)))
        if OBS in data:
            del keys[OBS]
            data[OBS] = add_list(data[OBS])
            for j in data[OBS]:
                test_list.append({**keys, **j})
        else:
            test_list.append(keys)
        test_list, df = __process_df(test_list, df)

    test_list, df = __process_df(test_list, df, is_end=True)

    return df


def _reading_group_data(dataset: Dict[str, Any]) -> pd.DataFrame:
    # Structure Specific Group Data
    test_list = []
    df = None
    dataset[GROUP] = add_list(dataset[GROUP])
    for data in dataset[GROUP]:
        test_list.append(dict(data.items()))
        test_list, df = __process_df(test_list, df)
    test_list, df = __process_df(test_list, df, is_end=True)

    cols_to_delete = [x for x in df.columns if ":type" in x]
    for x in cols_to_delete:
        del df[x]

    df = df.drop_duplicates(keep="first").reset_index(drop=True)

    return df


def _get_at_att_str(dataset: Dict[str, Any]) -> Dict[str, Any]:
    """Gets the elements of the dataset if it is Structure Specific Data."""
    return {k: dataset[k] for k in dataset if k not in EXCLUDED_ATTRIBUTES}


def _parse_structure_specific_data(
    dataset: Dict[str, Any], structure_info: Dict[str, Any]
) -> PandasDataset:
    attached_attributes = _get_at_att_str(dataset)

    df = pd.DataFrame()

    # Parsing data
    if SERIES in dataset:
        # Structure Specific Series
        df = _reading_str_series(dataset)
        if GROUP in dataset:
            df_group = _reading_group_data(dataset)
            common_columns = list(
                set(df.columns).intersection(set(df_group.columns))
            )
            df = pd.merge(df, df_group, on=common_columns, how="left")
    elif OBS in dataset:
        dataset[OBS] = add_list(dataset[OBS])
        # Structure Specific All dimensions
        df = pd.DataFrame(dataset[OBS]).replace(np.nan, "")

    urn = f"{structure_info['structure_type']}={structure_info['unique_id']}"
    action = dataset.get("action", "Information")
    action = ActionType(action)

    return PandasDataset(
        structure=urn, attributes=attached_attributes, data=df, action=action
    )



"""API client for interacting with the GDS (Global Discovery Service) service.

This module provides classes and utilities to interact with the GDS service,
allowing retrieval of metadata such as agency information in SDMX-JSON format.

Exports: GdsClient: A synchronous client for retrieving metadata from
the GDS.
"""

from typing import Any, Literal, Optional, Sequence

from msgspec.json import decode

from pysdmx.api.qb.gds import GdsQuery, GdsType
from pysdmx.api.qb.service import GdsAsyncRestService, GdsRestService
from pysdmx.api.qb.util import REST_ALL
from pysdmx.io.json.gds.reader import deserializers as gds_readers
from pysdmx.io.serde import Deserializer
from pysdmx.model import Agency
from pysdmx.model.gds import (
    GdsCatalog,
    GdsSdmxApi,
    GdsService,
    GdsUrnResolver,
)

GDS_BASE_ENDPOINT = "https://gds.sdmx.io/"

READERS = gds_readers


class __BaseGdsClient:
    def __init__(
        self,
        api_endpoint: str = GDS_BASE_ENDPOINT,
        pem: Optional[str] = None,
    ):
        """Instantiate a new client against the target endpoint."""
        if api_endpoint.endswith("/"):
            api_endpoint = api_endpoint[0:-1]
        self.api_endpoint = api_endpoint
        self.reader = READERS

    def _out(self, response: bytes, typ: Deserializer, *params: Any) -> Any:
        return decode(response, type=typ).to_model(*params)

    def _agencies_q(self, agency: str) -> GdsQuery:
        return GdsQuery(artefact_type=GdsType.GDS_AGENCY, resource_id=agency)

    def _catalogs_q(
        self,
        agency: str,
        resource: str = REST_ALL,
        version: str = REST_ALL,
        resource_type: Optional[Literal["data", "metadata"]] = None,
        message_format: Optional[Literal["json", "csv", "xml"]] = None,
        api_version: Optional[str] = None,
        detail: Optional[Literal["full", "raw"]] = None,
        references: Optional[Literal["none", "children"]] = None,
    ) -> GdsQuery:
        return GdsQuery(
            artefact_type=GdsType.GDS_CATALOG,
            agency=agency,
            resource_id=resource,
            version=version,
            resource_type=resource_type,
            message_format=message_format,
            api_version=api_version,
            detail=detail,
            references=references,
        )

    def _sdmx_api_q(self, id: str = REST_ALL) -> GdsQuery:
        return GdsQuery(artefact_type=GdsType.GDS_SDMX_API, resource_id=id)

    def _services_q(
        self, agency: str, resource: str = REST_ALL, version: str = REST_ALL
    ) -> GdsQuery:
        return GdsQuery(
            artefact_type=GdsType.GDS_SERVICE,
            agency=agency,
            resource_id=resource,
            version=version,
        )

    def _urn_resolver_q(self, urn: str) -> GdsQuery:
        return GdsQuery(
            artefact_type=GdsType.GDS_URN_RESOLVER, resource_id=urn
        )


class GdsClient(__BaseGdsClient):
    """A client to be used to retrieve metadata from the GDS.

    With this client, metadata will be retrieved in a synchronous fashion.
    """

    def __init__(
        self,
        api_endpoint: str = GDS_BASE_ENDPOINT,
        pem: Optional[str] = None,
    ) -> None:
        """Instantiate a new client against the target endpoint.

        Args:
            api_endpoint: The endpoint of the targeted service.
            pem: In case the service exposed a certificate created
                by an unknown certificate authority, you can pass
                a pem file for this authority using this parameter.
        """
        super().__init__(api_endpoint, pem)
        self.__service = GdsRestService(
            self.api_endpoint,
            pem=pem,
            timeout=10.0,
        )

    def __fetch(
        self,
        query: GdsQuery,
    ) -> bytes:
        """Fetch the requested metadata from the GDS service."""
        return self.__service.gds(query)

    def get_agencies(self, agency: str) -> Sequence[Agency]:
        """Get the list of agencies for the supplied name.

        Args:
            agency: The agency maintaining the agency scheme from
                which sub-agencies must be returned.

        Returns:
            The requested list of agencies.
        """
        query = super()._agencies_q(agency)
        out = self.__fetch(query)
        agencies = super()._out(out, self.reader.agencies)
        return agencies

    def get_catalogs(
        self,
        catalog: str,
        resource: str = REST_ALL,
        version: str = REST_ALL,
        resource_type: Optional[Literal["data", "metadata"]] = None,
        message_format: Optional[Literal["json", "csv", "xml"]] = None,
        api_version: Optional[str] = None,
        detail: Optional[Literal["full", "raw"]] = None,
        references: Optional[Literal["none", "children"]] = None,
    ) -> Sequence[GdsCatalog]:
        """Get the list of catalogs for the supplied parameters.

        Args:
            catalog: The agency maintaining the catalog.
            resource: The resource ID(s) to query. Defaults to '*'.
            version: The version(s) of the resource. Defaults to '*'.
            resource_type: Filters the endpoints that support
             the requested resource type (eg, 'data', 'metadata')
            message_format: Filters the endpoints that support any
              of the requested message formats.
            api_version: Filters the endpoints that is in a
              specific SDMX API version.
              Multiple values separated by commas are possible.
              By default (if nothing is sent) it returns everything.
            detail: The amount of information to be returned.
              If detail=full: All available information for all artefacts
              should be returned.
              If detail=raw: Any nested service will be referenced.
            references: Instructs the web service to return
              (or not) the artefacts referenced by the
              artefact to be returned.
              If references=none: No referenced artefacts will be returned.
              If references=children: Returns the artefacts
              referenced by the artefact to be returned.

        Returns:
            A list of GdsCatalog objects.
        """
        query = super()._catalogs_q(
            catalog,
            resource,
            version,
            resource_type,
            message_format,
            api_version,
            detail,
            references,
        )
        response = self.__fetch(query)
        catalogs = super()._out(response, self.reader.catalogs)
        return catalogs

    def get_sdmx_apis(
        self, api_version: str = REST_ALL
    ) -> Sequence[GdsSdmxApi]:
        """Get the list of SDMX API versions.

        Args:
            api_version: The version of the SDMX API to be returned.
              Defaults to '*'.
        """
        query = super()._sdmx_api_q(api_version)
        response = self.__fetch(query)
        sdmx_api = super()._out(response, self.reader.sdmx_api)
        return sdmx_api

    def get_services(
        self, service: str, resource: str = REST_ALL, version: str = REST_ALL
    ) -> Sequence[GdsService]:
        """Get the list of services for the supplied parameters.

        Args:
            service: The agency maintaining the service.
            resource: The resource ID(s) to query. Defaults to '*'.
            version: The version(s) of the resource. Defaults to '*'.

        Returns:
            A list of GdsService objects.
        """
        query = super()._services_q(service, resource, version)
        response = self.__fetch(query)
        services = super()._out(response, self.reader.services)
        return services

    def get_urn_resolver(self, urn: str) -> GdsUrnResolver:
        """Resolve a URN to its corresponding resource.

        Args:
            urn: The URN to resolve.

        Returns:
            A GdsUrnResolver object with the resolved information.
        """
        query = super()._urn_resolver_q(urn)
        response = self.__fetch(query)
        urn_resolution = super()._out(response, self.reader.urn_resolver)
        return urn_resolution


class AsyncGdsClient(__BaseGdsClient):
    """A client to be used to retrieve metadata from the GDS asynchronously."""

    def __init__(
        self,
        api_endpoint: str = GDS_BASE_ENDPOINT,
        pem: Optional[str] = None,
    ) -> None:
        """Instantiate a new client against the target endpoint.

        Args:
            api_endpoint: The endpoint of the targeted service.
            pem: PEM file for unknown certificate authorities.
        """
        super().__init__(api_endpoint, pem)
        self.__service = GdsAsyncRestService(
            self.api_endpoint,
            pem=pem,
            timeout=10.0,
        )

    async def __fetch(self, query: GdsQuery) -> bytes:
        """Fetch the requested metadata from the GDS service asynchronously."""
        return await self.__service.gds(query)

    async def get_agencies(self, agency: str) -> Sequence[Agency]:
        """Get the list of agencies for the supplied name asynchronously.

        Args:
            agency: The agency maintaining the agency scheme.

        Returns:
            The requested list of agencies.
        """
        query = super(AsyncGdsClient, self)._agencies_q(agency)
        out = await self.__fetch(query)
        agencies = super(AsyncGdsClient, self)._out(out, self.reader.agencies)
        return agencies

    async def get_catalogs(
        self,
        catalog: str,
        resource: str = REST_ALL,
        version: str = REST_ALL,
        resource_type: Optional[Literal["data", "metadata"]] = None,
        message_format: Optional[Literal["json", "csv", "xml"]] = None,
        api_version: Optional[str] = None,
        detail: Optional[Literal["full", "raw"]] = None,
        references: Optional[Literal["none", "children"]] = None,
    ) -> Sequence[GdsCatalog]:
        """Get the list of catalogs for the supplied params asynchronously."""
        query = super()._catalogs_q(
            catalog,
            resource,
            version,
            resource_type,
            message_format,
            api_version,
            detail,
            references,
        )
        response = await self.__fetch(query)
        catalogs = super()._out(response, self.reader.catalogs)
        return catalogs

    async def get_sdmx_apis(
        self, api_version: str = REST_ALL
    ) -> Sequence[GdsSdmxApi]:
        """Get the list of SDMX API versions asynchronously."""
        query = super()._sdmx_api_q(api_version)
        response = await self.__fetch(query)
        sdmx_api = super()._out(response, self.reader.sdmx_api)
        return sdmx_api

    async def get_services(
        self, service: str, resource: str = REST_ALL, version: str = REST_ALL
    ) -> Sequence[GdsService]:
        """Get a list of services for the supplied params asynchronously."""
        query = super()._services_q(service, resource, version)
        response = await self.__fetch(query)
        services = super()._out(response, self.reader.services)
        return services

    async def get_urn_resolver(self, urn: str) -> GdsUrnResolver:
        """Resolve a URN to its corresponding resource asynchronously."""
        query = super()._urn_resolver_q(urn)
        response = await self.__fetch(query)
        urn_resolution = super()._out(response, self.reader.urn_resolver)
        return urn_resolution



"""Message module.

This module contains the enumeration for the different types of messages that
can be written. It also contains the Header and Message classes that are used
to create the SDMX messages.

Classes:
    Header: Header for the SDMX messages.
    ActionType: Enumeration for the different types of actions that can be
        written.
    Message: Class that holds the content of the SDMX message.
    SubmissionResult: Class that represents the result of a submission.
"""

import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from msgspec import Struct

from pysdmx.errors import Invalid, NotFound
from pysdmx.model.__base import ItemScheme, MaintainableArtefact, Organisation
from pysdmx.model.category import Categorisation, CategoryScheme
from pysdmx.model.code import Codelist, Hierarchy, HierarchyAssociation
from pysdmx.model.concept import ConceptScheme
from pysdmx.model.dataflow import (
    Dataflow,
    DataStructureDefinition,
    ProvisionAgreement,
)
from pysdmx.model.dataset import ActionType, Dataset
from pysdmx.model.map import (
    MultiRepresentationMap,
    RepresentationMap,
    StructureMap,
)
from pysdmx.model.metadata import MetadataReport
from pysdmx.model.organisation import AgencyScheme, DataProviderScheme
from pysdmx.model.submission import SubmissionResult
from pysdmx.model.vtl import (
    CustomTypeScheme,
    NamePersonalisationScheme,
    RulesetScheme,
    TransformationScheme,
    UserDefinedOperatorScheme,
    VtlMappingScheme,
)


class Header(Struct, repr_omit_defaults=True, kw_only=True):
    """Header for the SDMX messages.

    Represents the Header of an SDMX message, containing metadata about the
    message such as the sender, receiver, and other relevant information.

    Attributes:
        id: Unique identifier for the message. (default: generated UUID)
        test: Indicates if the message is a test message. (default: False)
        prepared: Timestamp when the message was prepared.
          (default: current UTC time)
        sender: Organisation that sent the message.
          (default: Organisation with id "ZZZ")
        receiver: Optional Organisation that received the message.
          (default: None)
        source: Optional source of the message. (default: None)
        dataset_action: Optional action for the dataset
          (only for SDMX Data messages). (default: None)
        structure: Dimension at observation mapping
          (dict with short URN as key and Dimension ID as value)
          (only for SDMX Data Messages)
          (Overridden by dimension_at_observation argument in writers).
          (default: None)
        dataset_id: DatasetID defined at SDMX-ML
          (only for SDMX-ML Data messages). (default: None)
    """

    id: str = str(uuid.uuid4())
    test: bool = False
    prepared: datetime = datetime.now(timezone.utc)
    sender: Organisation = Organisation(id="ZZZ")
    receiver: Optional[Organisation] = None
    source: Optional[str] = None
    dataset_action: Optional[ActionType] = None
    structure: Optional[Dict[str, str]] = None
    dataset_id: Optional[str] = None

    def __post_init__(self) -> None:
        """Header post-initialization."""
        if isinstance(self.sender, str):
            self.sender = Organisation(id=self.sender)

        if isinstance(self.receiver, str):
            self.receiver = Organisation(id=self.receiver)

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"


class StructureMessage(Struct, repr_omit_defaults=True, frozen=True):
    """Message class holds the content of an SDMX Structure Message.

    Attributes:
        header: The header of the SDMX message.
        structures: Sequence of MaintainableArtefact objects.
          They represent the contents of a Structure Message.
    """

    header: Optional[Header] = None
    structures: Optional[Sequence[MaintainableArtefact]] = None

    def __post_init__(self) -> None:
        """Checks if the content is valid."""
        if self.structures is not None:
            for obj_ in self.structures:
                if not isinstance(obj_, (MaintainableArtefact)):
                    raise Invalid(
                        f"Invalid structure: {type(obj_).__name__} ",
                        "Check the docs on structures.",
                    )

    def __str__(self) -> str:
        """Custom string representation with detailed structure counts."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            if attr in ["data", "structures"] and value:
                # Count occurrences of each class in structures
                class_counts: Dict[str, int] = {}
                for obj in value:
                    class_name = obj.__class__.__name__
                    class_counts[class_name] = (
                        class_counts.get(class_name, 0) + 1
                    )

                # Format the counts
                value = ", ".join(
                    f"{count} {class_name.lower()}"
                    for class_name, count in class_counts.items()
                )

            # Handle sequences and omit empty ones
            if (
                isinstance(value, Sequence)
                and not isinstance(value, str)
                and not value
            ):
                continue

            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # Omit empty sequences
            if isinstance(value, (list, tuple, set)) and not value:
                continue
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"

    def __get_elements(self, type_: Type[Any]) -> List[Any]:
        """Returns a list of elements of a specific type."""
        if self.structures is None:
            raise NotFound(
                f"No {type_.__name__} found in message.",
            )
        structures = []
        for element in self.structures:
            if isinstance(element, type_):
                structures.append(element)
        return structures

    def __get_enumerations(
        self, type_: Type[Any], is_vl: bool = False
    ) -> List[Any]:
        """Returns a list of elements of a specific type."""
        enums = self.__get_elements(type_)
        t = "valuelist" if is_vl else "codelist"
        return [e for e in enums if e.sdmx_type == t]

    def __get_single_structure(
        self,
        type_: Type[Union[ItemScheme, DataStructureDefinition, Dataflow]],
        short_urn: str,
    ) -> Any:
        """Returns a specific element from content."""
        if self.structures is None:
            raise NotFound(
                f"No {type_.__name__} found in message.",
                "Could not find any Structures in this message.",
            )
        for structure in self.structures:
            if structure.short_urn == short_urn:
                return structure

        raise NotFound(
            f"No {type_.__name__} with Short URN {short_urn} found in message",
            "Could not find the requested element.",
        )

    def get_agency_schemes(self) -> List[AgencyScheme]:
        """Returns the AgencySchemes."""
        return self.__get_elements(AgencyScheme)

    def get_codelists(self) -> List[Codelist]:
        """Returns the Codelists."""
        return self.__get_enumerations(Codelist, False)

    def get_concept_schemes(self) -> List[ConceptScheme]:
        """Returns the Concept Schemes."""
        return self.__get_elements(ConceptScheme)

    def get_data_structure_definitions(
        self,
    ) -> List[DataStructureDefinition]:
        """Returns the DataStructureDefinitions."""
        return self.__get_elements(DataStructureDefinition)

    def get_dataflows(self) -> List[Dataflow]:
        """Returns the Dataflows."""
        return self.__get_elements(Dataflow)

    def get_organisation_scheme(self, short_urn: str) -> AgencyScheme:
        """Returns a specific OrganisationScheme."""
        return self.__get_single_structure(AgencyScheme, short_urn)

    def get_codelist(self, short_urn: str) -> Codelist:
        """Returns a specific Codelist."""
        return self.__get_single_structure(Codelist, short_urn)

    def get_concept_scheme(self, short_urn: str) -> ConceptScheme:
        """Returns a specific Concept Scheme."""
        return self.__get_single_structure(ConceptScheme, short_urn)

    def get_data_structure_definition(
        self, short_urn: str
    ) -> DataStructureDefinition:
        """Returns a specific DataStructureDefinition."""
        return self.__get_single_structure(DataStructureDefinition, short_urn)

    def get_dataflow(self, short_urn: str) -> Dataflow:
        """Returns a specific Dataflow."""
        return self.__get_single_structure(Dataflow, short_urn)

    def get_transformation_schemes(self) -> List[TransformationScheme]:
        """Returns the TransformationSchemes."""
        return self.__get_elements(TransformationScheme)

    def get_user_defined_operator_schemes(
        self,
    ) -> List[UserDefinedOperatorScheme]:
        """Returns the UserDefinedOperatorSchemes."""
        return self.__get_elements(UserDefinedOperatorScheme)

    def get_ruleset_schemes(self) -> List[RulesetScheme]:
        """Returns the RulesetSchemes."""
        return self.__get_elements(RulesetScheme)

    def get_category_schemes(self) -> List[CategoryScheme]:
        """Returns the CategorySchemes."""
        return self.__get_elements(CategoryScheme)

    def get_value_lists(self) -> List[Codelist]:
        """Returns the Codelists."""
        return self.__get_enumerations(Codelist, True)

    def get_hierarchies(self) -> List[Hierarchy]:
        """Returns the HierarchyCodelists."""
        return self.__get_elements(Hierarchy)

    def get_hierarchy_associations(self) -> List[HierarchyAssociation]:
        """Returns the HierarchyAssociations."""
        return self.__get_elements(HierarchyAssociation)

    def get_data_provider_schemes(self) -> List[DataProviderScheme]:
        """Returns the DataProviderSchemes."""
        return self.__get_elements(DataProviderScheme)

    def get_provision_agreements(self) -> List[ProvisionAgreement]:
        """Returns the ProvisionAgreements."""
        return self.__get_elements(ProvisionAgreement)

    def get_structure_maps(self) -> List[StructureMap]:
        """Returns the StructureMaps."""
        return self.__get_elements(StructureMap)

    def get_representation_maps(
        self,
    ) -> List[Union[MultiRepresentationMap, RepresentationMap]]:
        """Returns the RepresentationMaps."""
        out = []
        out.extend(self.__get_elements(RepresentationMap))
        out.extend(self.__get_elements(MultiRepresentationMap))
        return out

    def get_categorisations(self) -> List[Categorisation]:
        """Returns the Categorisations."""
        return self.__get_elements(Categorisation)

    def get_custom_type_schemes(self) -> List[CustomTypeScheme]:
        """Returns the CustomType Schemes."""
        return self.__get_elements(CustomTypeScheme)

    def get_vtl_mapping_schemes(self) -> List[VtlMappingScheme]:
        """Returns the VTL Mapping Schemes."""
        return self.__get_elements(VtlMappingScheme)

    def get_name_personalisation_schemes(
        self,
    ) -> List[NamePersonalisationScheme]:
        """Returns the NamePersonalisationSchemes."""
        return self.__get_elements(NamePersonalisationScheme)


class MetadataMessage(Struct, frozen=True):
    """Message class holds the content of an SDMX Reference Metadata Message.

    Attributes:
        header: The header of the SDMX message.
        reports: Sequence of metadata reports.
    """

    header: Optional[Header] = None
    reports: Optional[Sequence[MetadataReport]] = None

    def get_reports(self) -> Sequence[MetadataReport]:
        """Returns the metadata reports."""
        if self.reports:
            return self.reports
        else:
            raise NotFound("No metadata reports werefound in the message.")


class Message(StructureMessage, frozen=True):
    """Message class holds the content of SDMX Message.

    Attributes:
        header: The header of the SDMX message.
        structures: Sequence of MaintainableArtefact objects.
        data: Sequence of Dataset objects. They represent the contents of a
           SDMX Data Message in any format.
        submission: Sequence of SubmissionResult objects. They represent the
              contents of a SDMX Submission Message.
    """

    data: Optional[Sequence[Dataset]] = None
    submission: Optional[Sequence[SubmissionResult]] = None

    def __post_init__(self) -> None:
        """Checks if the content is valid."""
        super().__post_init__()
        if self.data is not None:
            for data_value in self.data:
                if not isinstance(data_value, Dataset):
                    raise Invalid(
                        f"Invalid data type: "
                        f"{type(data_value).__name__} "
                        f"for Data Message, requires a Dataset object.",
                        "Check the docs for the proper structure on data.",
                    )

    def get_datasets(self) -> Sequence[Dataset]:
        """Returns the Datasets."""
        if self.data is not None:
            return self.data
        raise NotFound(
            "No Datasets found in data.",
            "Could not find any Datasets in content.",
        )

    def get_dataset(self, short_urn: str) -> Dataset:
        """Returns a specific Dataset."""
        if self.data is not None:
            for dataset in self.data:
                if dataset.short_urn == short_urn:
                    return dataset
        raise NotFound(
            f"No Dataset with Short URN {short_urn} found in data.",
            "Could not find the requested Dataset.",
        )



"""Collection of GDS-JSON schemas for GDS catalogs."""

from typing import Any, List, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.gds.messages.agencies import JsonAgency
from pysdmx.io.json.gds.messages.services import JsonService
from pysdmx.model import Agency
from pysdmx.model.gds import (
    GdsCatalog,
    GdsEndpoint,
    GdsService,
    GdsServiceReference,
)


class JsonCatalog(Struct, frozen=True):
    """GDS-JSON payload for a catalog scheme."""

    agencyID: str
    id: str
    version: str
    name: str
    urn: str
    endpoints: Optional[List[GdsEndpoint]] = None
    serviceRefs: Optional[List[GdsServiceReference]] = None

    def to_model(
        self, agencies: List[Agency], services: List[GdsService]
    ) -> GdsCatalog:
        """Converts the payload to a GDS Catalog."""
        agency = next(
            (a for a in agencies if a.id == self.agencyID),
            self.agencyID,
        )

        catalog_services: Any = services
        if self.serviceRefs:
            catalog_services = catalog_services or []
            catalog_services.extend(
                ref
                for ref in self.serviceRefs
                if ref.short_urn not in {s.short_urn for s in catalog_services}
            )

        return GdsCatalog(
            id=self.id,
            version=self.version,
            name=self.name,
            urn=self.urn,
            agency=agency,
            endpoints=self.endpoints,
            services=catalog_services,
        )


class JsonStructures(Struct, frozen=True):
    """Intermediate structure for 'structures' field."""

    catalogs: Sequence[JsonCatalog]
    agencies: Optional[Sequence[JsonAgency]] = None
    services: Optional[List[JsonService]] = None


class JsonCatalogMessage(Struct, frozen=True):
    """GDS-JSON payload for /catalog queries."""

    structures: JsonStructures

    def to_model(self) -> Sequence[GdsCatalog]:
        """Returns a list of GdsCatalog objects."""
        agencies = [a.to_model() for a in (self.structures.agencies or [])]
        services = [s.to_model() for s in (self.structures.services or [])]

        return [
            c.to_model(agencies=agencies, services=services)
            for c in self.structures.catalogs
        ]



"""Model validations for VTL models."""

from typing import Union

from pysdmx.__extras_check import __check_vtl_extra
from pysdmx.errors import Invalid
from pysdmx.model import (
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlScheme,
)
from pysdmx.model.__base import Item


def model_validations(model_obj: Union[VtlScheme, Item]) -> None:
    """Validation checks for VTL objects.

    This method performs validation checks on VTLScheme objects ,
    such as TransformationScheme, RulesetScheme,
    UserDefinedOperatorScheme, etc. or its items,
    like Ruleset, Transformation, UserDefinedOperator, etc.

    It raises an Invalid exception if the model object is not valid.

    The model validation checks on items:

    * Ruleset validation

        * A single RulesetDefinition (define ...) is valid in a Ruleset

        * Ruleset type matches the definition (hierarchical or datapoint)

        * Ruleset scope matches the signature (variable or valuedomain)

    * User Defined Operator validation

        * A single OperatorDefinition (define operator ...)
          is valid in a User Defined Operator

    * Transformation validation

        * Checks if the transformation is valid

        * Checks a single assignment is present in the expression

    The model validation checks on VTLScheme:

    * RulesetScheme validation

        * Checks if it contains at least one Ruleset

        * Checks if all items in the scheme are Ruleset

    * UserDefinedOperatorScheme validation

        * Checks if it contains at least one UserDefinedOperator

        * Checks if all items in the scheme are UserDefinedOperator

    * TransformationScheme validation

        * Checks if it contains at least one Transformation

        * Checks if all items in the scheme are Transformation

        * Checks the referenced RulesetSchemes (if any)
          and UserDefinedOperatorSchemes (if any)

    Args:
        model_obj: A VTLScheme or Item object.
    raises:
        Invalid: Invalid model object if the model object is not valid.
    """
    # We add here the check for vtl extra and add the local imports to
    # prevent unhandled ImportError
    __check_vtl_extra()

    if isinstance(model_obj, Ruleset):
        from pysdmx.toolkit.vtl._validations import (
            _ruleset_validation,
        )

        _ruleset_validation(model_obj)
    elif isinstance(model_obj, RulesetScheme):
        from pysdmx.toolkit.vtl._validations import (
            _ruleset_scheme_validations,
        )

        _ruleset_scheme_validations(model_obj)
    elif isinstance(model_obj, UserDefinedOperator):
        from pysdmx.toolkit.vtl._validations import (
            _user_defined_operator_validation,
        )

        _user_defined_operator_validation(model_obj)
    elif isinstance(model_obj, UserDefinedOperatorScheme):
        from pysdmx.toolkit.vtl._validations import (
            _user_defined_operator_scheme_validations,
        )

        _user_defined_operator_scheme_validations(model_obj)
    elif isinstance(model_obj, Transformation):
        from pysdmx.toolkit.vtl._validations import (
            _transformation_validations,
        )

        _transformation_validations(model_obj)
    elif isinstance(model_obj, TransformationScheme):
        from pysdmx.toolkit.vtl._validations import (
            _transformation_scheme_validations,
        )

        _transformation_scheme_validations(model_obj)
    else:
        raise Invalid("Invalid model object")



"""VTL toolkit for PySDMX."""

from pysdmx.toolkit.vtl.script_generation import generate_vtl_script
from pysdmx.toolkit.vtl.validation import model_validations

__all__ = ["model_validations", "generate_vtl_script"]



"""SDMX XML 3.0 StructureSpecificData reader module."""

from typing import Sequence

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.__data_aux import (
    get_data_objects,
)
from pysdmx.io.xml.__parse_xml import parse_xml
from pysdmx.io.xml.__ss_aux_reader import (
    _parse_structure_specific_data,
)
from pysdmx.io.xml.__tokens import (
    STR_REF,
    STR_SPE,
)


def read(input_str: str, validate: bool = True) -> Sequence[PandasDataset]:
    """Reads an SDMX-ML 3.0 and returns a Sequence of Datasets.

    Args:
        input_str: SDMX-ML data to read.
        validate: If True, the XML data will be validated against the XSD.
    """
    dict_info = parse_xml(input_str, validate=validate)
    if STR_SPE not in dict_info:
        raise Invalid(
            "This SDMX document is not an SDMX-ML StructureSpecificData."
        )
    dataset_info, str_info = get_data_objects(dict_info[STR_SPE])
    datasets = []
    for dataset in dataset_info:
        ds = _parse_structure_specific_data(
            dataset, str_info[dataset[STR_REF]]
        )
        datasets.append(ds)
    return datasets



"""A module to represent a Submission Result."""

from msgspec import Struct


class SubmissionResult(Struct, frozen=True):
    """A class to represent a Submission Result."""

    action: str
    short_urn: str
    status: str

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"



"""pysdmx.io.writer.

Common data and structure writer for SDMX objects.
"""

from typing import Any, Optional, Sequence

from pysdmx.errors import Invalid
from pysdmx.io.format import Format
from pysdmx.model.dataset import Dataset

WRITERS = {
    Format.DATA_SDMX_CSV_1_0_0: "pysdmx.io.csv.sdmx10.writer",
    Format.DATA_SDMX_CSV_2_0_0: "pysdmx.io.csv.sdmx20.writer",
    Format.DATA_SDMX_ML_2_1_GEN: "pysdmx.io.xml.sdmx21.writer.generic",
    Format.DATA_SDMX_ML_2_1_STR: "pysdmx.io.xml.sdmx21.writer."
    "structure_specific",
    Format.STRUCTURE_SDMX_ML_2_1: "pysdmx.io.xml.sdmx21.writer.structure",
    Format.DATA_SDMX_ML_3_0: "pysdmx.io.xml.sdmx30.writer."
    "structure_specific",
    Format.STRUCTURE_SDMX_ML_3_0: "pysdmx.io.xml.sdmx30.writer.structure",
    Format.DATA_SDMX_ML_3_1: "pysdmx.io.xml.sdmx31.writer."
    "structure_specific",
    Format.STRUCTURE_SDMX_ML_3_1: "pysdmx.io.xml.sdmx31.writer.structure",
}

STRUCTURE_WRITERS = (
    Format.STRUCTURE_SDMX_ML_2_1,
    Format.STRUCTURE_SDMX_ML_3_0,
    Format.STRUCTURE_SDMX_ML_3_1,
)


def write_sdmx(
    sdmx_objects: Any,
    sdmx_format: Format,
    output_path: str = "",
    **kwargs: Any,
) -> Optional[str]:
    """Writes any SDMX object (or list of them) to any supported SDMX format.

    See the :ref:`formats available <io-writer-formats-supported>`

    .. important::
        To use the pysdmx.io data functionalities, you need to
        install the `pysdmx[data]` extra.

        For SDMX-ML support, you also need to install the `pysdmx[xml]` extra.

        Check the :ref:`installation guide <installation>`
        for more information.

    .. important::
        To write SDMX-ML Generic or Series messages, the PandasDataset
        requires to have its structure defined as a
        :class:`Schema <pysdmx.model.dataflow.Schema>`.

    Args:
        sdmx_objects: Model objects to write, including PandasDataset,
            DataStructure, Dataflow, ConceptScheme, etc.
        sdmx_format: The pysdmx.io.Format to write to, e.g.,
            Format.DATA_SDMX_ML_3_0.
        output_path: The path to save the file. If empty, returns a string.
        **kwargs: Additional keyword arguments (see below).

    Keyword Args:
        prettyprint: Whether to pretty-print the output (default: True)
          (only for SDMX-ML).
        header: Custom :class:`Header <pysdmx.model.message.Header>` to
          include in the SDMX Message (only for SDMX-ML)
        dimension_at_observation: Mapping for dimension at observation
          (only for SDMX-ML Data formats). This is a dictionary where
          the keys are short URNs and the values are the dimension IDs
          that should be used as the dimension at observation for that
          structure in the output. For example,
          ``{"Dataflow=MD:TEST_MD(1.0)": "TIME_PERIOD"}``.
          Overrides the header.structure
          (if a custom header is provided).

    Returns:
        A serialised string if output_path is an empty string, otherwise None.

    Raises:
        Invalid: If the file is empty or the format is not supported.
    """
    if sdmx_format not in WRITERS:
        raise Invalid(
            f"No writer found for format: {sdmx_format}. "
            f"Check the docs for supported formats."
        )

    sdmx_objects = (
        sdmx_objects if isinstance(sdmx_objects, Sequence) else [sdmx_objects]
    )

    module = __import__(WRITERS[sdmx_format], fromlist=["write"])
    writer = module.write

    is_structure = sdmx_format in STRUCTURE_WRITERS
    is_xml = "xml" in WRITERS[sdmx_format]
    key = "structures" if is_structure else "datasets"
    value = sdmx_objects if isinstance(sdmx_objects, list) else [sdmx_objects]

    if is_structure and any(isinstance(x, Dataset) for x in value):
        raise Invalid(
            "Datasets cannot be written to structure formats. "
            "Use data formats instead."
        )
    elif not is_structure and not all(isinstance(x, Dataset) for x in value):
        raise Invalid(
            "Only Datasets can be written to data formats. "
            "Use structure formats for other SDMX objects."
        )

    args = {
        key: value,
        "output_path": output_path,
        **(
            {
                "prettyprint": kwargs.get("prettyprint"),
                "header": kwargs.get("header"),
            }
            if is_xml
            else {}
        ),
        **(
            {
                "dimension_at_observation": kwargs.get(
                    "dimension_at_observation"
                )
            }
            if is_xml and not is_structure
            else {}
        ),
    }
    args = {k: v for k, v in args.items() if v is not None}

    return writer(**args)



"""Build GDS-REST structure queries."""

from enum import Enum
from typing import Literal, Optional, Sequence, Union

import msgspec

from pysdmx.api.qb.util import (
    REST_ALL,
    REST_LATEST,
)
from pysdmx.errors import Invalid


class GdsType(Enum):
    """The type of GDS metadata to be returned."""

    GDS_AGENCY = "agency"
    GDS_CATALOG = "catalog"
    GDS_SDMX_API = "sdmxapi"
    GDS_SERVICE = "service"
    GDS_URN_RESOLVER = "urn_resolver"
    ALL = REST_ALL
    LATEST = REST_LATEST


_RESOURCES = {
    GdsType.GDS_AGENCY,
    GdsType.GDS_CATALOG,
    GdsType.GDS_SDMX_API,
    GdsType.GDS_SERVICE,
    GdsType.GDS_URN_RESOLVER,
}


class GdsQuery(msgspec.Struct, frozen=True, omit_defaults=True):
    """A query for base GDS metadata.

    Attributes:
        artefact_type: The type of GDS metadata to be returned.
        agency: The agency (or agencies) maintaining the artefact(s)
            to be returned.
        resource_id: The resource ID(s) to query. Defaults to '*'.
        version: The version(s) of the resource. Defaults to '*'.
        resource_type: Filters the endpoints that support
          the requested resource type (eg, 'data', 'metadata').
        message_format: Filters the endpoints that support any
          of the requested message formats.

        api_version: Filters the endpoints that is in a
          specific SDMX API version.
          Multiple values separated by commas are possible.
          By default (if nothing is sent) it returns everything.

        detail: The amount of information to be returned.
            Option full: All available information for all artefacts
              should be returned.
            Option raw: Any nested service will be referenced.

        references: Instructs the web service to return
          (or not) the artefacts referenced by the
          artefact to be returned.

            Option none: No referenced artefacts will be returned.
            Option children: Returns the artefacts
              referenced by the artefact to be returned.
    """

    artefact_type: GdsType
    agency: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Optional[str] = None
    resource_type: Optional[Literal["data", "metadata"]] = None
    message_format: Optional[Literal["json", "csv", "xml"]] = None
    api_version: Optional[str] = None
    detail: Optional[Literal["full", "raw"]] = None
    references: Optional[Literal["none", "children"]] = None

    def validate(self) -> None:
        """Validate the query."""
        try:
            decoder.decode(encoder.encode(self))
        except msgspec.DecodeError as err:
            raise Invalid("Invalid Structure Query", str(err)) from err

    def _get_base_url(self) -> str:
        """The URL for the query in the GDS-REST API."""
        self.__validate_query()
        return self.__create_query()

    def get_url(self) -> str:
        """The URL for the query in the GDS-REST API."""
        base_url = self._get_base_url()
        params = []
        if self.resource_type:
            params.append(f"resource_type={self.resource_type}")
        if self.message_format:
            params.append(f"message_format={self.message_format}")
        if self.api_version:
            params.append(f"api_version={self.api_version}")
        if self.detail:
            params.append(f"detail={self.detail}")
        if self.references:
            params.append(f"references={self.references}")
        query_string = f"/?{'&'.join(params)}" if params else ""
        return f"{base_url}{query_string}"

    def __validate_query(self) -> None:
        self.validate()
        self.__check_artefact_type(self.artefact_type)

    def __check_artefact_type(self, atyp: GdsType) -> None:
        if atyp not in _RESOURCES:
            raise Invalid(
                "Validation Error",
                f"{atyp} is not valid for GDS-REST.",
            )

    def __to_type_kw(self, val: GdsType) -> str:
        return val.value

    def __to_kw(self, val: str) -> str:
        return val

    def __to_kws(self, vals: Union[str, Sequence[str]]) -> str:
        vals = [vals] if isinstance(vals, str) else vals
        mapped = [self.__to_kw(v) for v in vals]
        sep = ","
        return sep.join(mapped)

    def __create_query(self) -> str:
        t = self.__to_type_kw(self.artefact_type)
        a = self.__to_kws(self.agency)
        r = self.__to_kws(self.resource_id)
        v = self.__to_kws(self.version) if self.version else REST_ALL

        if t in ("sdmxapi", "agency", "urn_resolver"):
            return f"/{t}/{r}"

        vu = f"/{v}" if v != REST_ALL else ""
        ru = f"/{r}{vu}" if vu or self.resource_id != REST_ALL else ""
        au = f"/{a}{ru}" if ru or self.agency != REST_ALL else ""

        return f"/{t}{au}"


decoder = msgspec.json.Decoder(GdsQuery)
encoder = msgspec.json.Encoder()


__all__ = [
    "GdsQuery",
]



"""Schemas for SDMX-JSON messages."""

from pysdmx.io.json.sdmxjson2.messages.agency import JsonAgencyMessage
from pysdmx.io.json.sdmxjson2.messages.category import (
    JsonCategorisationMessage,
    JsonCategorySchemeMessage,
)
from pysdmx.io.json.sdmxjson2.messages.code import (
    JsonCodelistMessage,
    JsonHierarchiesMessage,
    JsonHierarchyAssociationMessage,
    JsonHierarchyMessage,
)
from pysdmx.io.json.sdmxjson2.messages.concept import JsonConceptSchemeMessage
from pysdmx.io.json.sdmxjson2.messages.dataflow import (
    JsonDataflowMessage,
    JsonDataflowsMessage,
)
from pysdmx.io.json.sdmxjson2.messages.dsd import JsonDataStructuresMessage
from pysdmx.io.json.sdmxjson2.messages.map import (
    JsonMappingMessage,
    JsonRepresentationMapMessage,
    JsonRepresentationMapsMessage,
    JsonStructureMapsMessage,
)
from pysdmx.io.json.sdmxjson2.messages.pa import (
    JsonProvisionAgreementsMessage as JsonPAMessage,
)
from pysdmx.io.json.sdmxjson2.messages.provider import JsonProviderMessage
from pysdmx.io.json.sdmxjson2.messages.report import JsonMetadataMessage
from pysdmx.io.json.sdmxjson2.messages.schema import JsonSchemaMessage
from pysdmx.io.json.sdmxjson2.messages.structure import JsonStructureMessage
from pysdmx.io.json.sdmxjson2.messages.vtl import (
    JsonVtlTransformationsMessage as JsonTransfoMsg,
)

__all__ = [
    "JsonAgencyMessage",
    "JsonCategorisationMessage",
    "JsonCategorySchemeMessage",
    "JsonCodelistMessage",
    "JsonConceptSchemeMessage",
    "JsonDataflowMessage",
    "JsonDataflowsMessage",
    "JsonDataStructuresMessage",
    "JsonProviderMessage",
    "JsonPAMessage",
    "JsonSchemaMessage",
    "JsonHierarchyAssociationMessage",
    "JsonHierarchiesMessage",
    "JsonHierarchyMessage",
    "JsonMetadataMessage",
    "JsonMappingMessage",
    "JsonRepresentationMapMessage",
    "JsonRepresentationMapsMessage",
    "JsonStructureMapsMessage",
    "JsonStructureMessage",
    "JsonTransfoMsg",
]



from typing import NoReturn, Union

import httpx

from pysdmx import errors


def map_httpx_errors(
    e: Union[httpx.RequestError, httpx.HTTPStatusError],
) -> NoReturn:
    """Map httpx errors to pysdmx errors."""
    q = e.request.url
    if isinstance(e, httpx.HTTPStatusError):
        s = e.response.status_code
        t = e.response.text
        if s == 404:
            msg = (
                "The requested resource(s) could not be found in the "
                f"targeted service. The query was `{q}`"
            )
            raise errors.NotFound("Not found", msg) from e
        elif s < 500:
            msg = (
                f"The query returned a {s} error code. The query "
                f"was `{q}`. The error message was: `{t}`."
            )
            raise errors.Invalid(f"Client error {s}", msg) from e
        else:
            msg = (
                f"The service returned a {s} error code. The query "
                f"was `{q}`. The error message was: `{t}`."
            )
            raise errors.InternalError(f"Service error {s}", msg) from e
    else:
        msg = (
            f"There was an issue connecting to the targeted service. "
            f"The query was `{q}`. The error message was: `{e}`."
        )
        raise errors.Unavailable("Connection error", msg) from e



"""Collection of GDS-JSON schemas for GDS sdmx_api obj."""

from typing import Sequence

from msgspec import Struct

from pysdmx.model.gds import GdsSdmxApi


class JsonStructures(Struct, frozen=True):
    """Intermediate structure for 'structures' field."""

    sdmxApis: Sequence[GdsSdmxApi]


class JsonSdmxApiMessage(Struct, frozen=True):
    """GDS-JSON payload for /sdmxapi queries."""

    structures: JsonStructures

    def to_model(self) -> Sequence[GdsSdmxApi]:
        """Returns a list of GdsSdmxApi objects."""
        return self.structures.sdmxApis



# mypy: disable-error-code="union-attr"
"""Module for writing SDMX-ML 2.1 Generic data messages."""

from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import pandas as pd

from pysdmx.io.format import Format
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.__write_aux import (
    ABBR_GEN,
    ABBR_MSG,
    ALL_DIM,
    __escape_xml,
    __write_header,
    create_namespaces,
    get_end_message,
    get_structure,
)
from pysdmx.io.xml.__write_data_aux import (
    check_content_dataset,
    check_dimension_at_observation,
    get_codes,
    writing_validation,
)
from pysdmx.io.xml.config import CHUNKSIZE
from pysdmx.model.message import Header
from pysdmx.util import parse_short_urn


def __value(id: str, value: str) -> str:
    """Write a value tag."""
    return f"<{ABBR_GEN}:Value id={id!r} value={__escape_xml(str(value))!r}/>"


def __generate_obs_structure(
    dataset: PandasDataset,
) -> Tuple[List[str], str, List[str]]:
    """Generate the structure of the observations.

    First element is a list of
    the codes for Key, second element is the code for the value, and the third
    element is a list of the codes for the attributes.

    Args:
        dataset: PandasDataset. The dataset to generate the structure from.

    Returns:
        The structure of the observations
    """
    obs_structure: Tuple[List[str], str, List[str]] = (
        [],
        dataset.structure.components.measures[0].id,
        [],
    )
    for dim in dataset.structure.components.dimensions:
        obs_structure[0].append(dim.id)

    cols = dataset.data.columns

    for att in dataset.structure.components.attributes:
        if att.attachment_level == "O" and att.id in cols:
            obs_structure[2].append(att.id)
        elif (
            att.attachment_level is not None
            and att.attachment_level != "D"
            and att.id in cols
        ):
            obs_structure[0].append(att.id)

    return obs_structure


def __memory_optimization_writing(
    dataset: PandasDataset,
    obs_structure: Tuple[List[str], str, List[str]],
    prettyprint: bool,
) -> str:
    """Memory optimization for writing data."""
    outfile = ""
    length_ = len(dataset.data)
    if len(dataset.data) > CHUNKSIZE:
        previous = 0
        next_ = CHUNKSIZE
        while previous <= length_:
            # Sliding a window for efficient access to the data
            # and avoid memory issues
            outfile += __obs_processing(
                dataset.data.iloc[previous:next_], obs_structure, prettyprint
            )
            previous = next_
            next_ += CHUNKSIZE

            if next_ >= length_:
                outfile += __obs_processing(
                    dataset.data.iloc[previous:], obs_structure, prettyprint
                )
                previous = next_
    else:
        outfile += __obs_processing(dataset.data, obs_structure, prettyprint)

    return outfile


def __write_data_generic(
    datasets: Dict[str, PandasDataset],
    dim_mapping: Dict[str, str],
    prettyprint: bool = True,
) -> str:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        datasets: dict. Datasets to be written.
        dim_mapping: dict. URN-DimensionAtObservation mapping.
        prettyprint: bool. Prettyprint or not.

    Returns:
        The data in SDMX-ML 2.1 Generic format, as string.
    """
    outfile = ""

    for short_urn, dataset in datasets.items():
        writing_validation(dataset)
        dataset.data = dataset.data.fillna("").astype(str)
        outfile += __write_data_single_dataset(
            dataset=dataset,
            prettyprint=prettyprint,
            dim=dim_mapping[short_urn],
        )

    return outfile


def __write_data_single_dataset(
    dataset: PandasDataset,
    prettyprint: bool = True,
    dim: str = ALL_DIM,
) -> str:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        dataset: PandasDataset. Dataset to be written.
        prettyprint: bool. Prettyprint or not.
        dim: str. Dimension to be written.

    Returns:
        The data in SDMX-ML 2.1 Generic format, as string.
    """

    def __remove_optional_attributes_empty_data(str_to_check: str) -> str:
        """This function removes data when optional attributes are found."""
        for att in dataset.structure.components.attributes:
            if not att.required:
                to_replace = f'<{ABBR_GEN}:Value id={att.id!r} value=""/>'
                to_replace = f"{child3}{to_replace}{nl}"
                str_to_check = str_to_check.replace(to_replace, "")
        return str_to_check

    outfile = ""
    structure_urn = get_structure(dataset)
    id_structure = parse_short_urn(structure_urn).id
    dataset.data = dataset.data.fillna("").astype(str).replace("nan", "")

    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""

    # Datasets
    outfile += (
        f"{nl}{child1}<{ABBR_MSG}:DataSet "
        f"structureRef={id_structure!r} "
        f"action={dataset.action.value!r}>{nl}"
    )
    data = ""
    # Write attached attributes
    attached_attributes_str = []
    for k, v in dataset.attributes.items():
        attached_attributes_str.append(__value(k, v))

    if len(attached_attributes_str) > 0:
        data += f"{child2}<{ABBR_GEN}:Attributes>{nl}"
        for att in attached_attributes_str:
            data += f"{child3}{att}{nl}"
        data += f"{child2}</{ABBR_GEN}:Attributes>{nl}"

    if dim == ALL_DIM:
        obs_structure = __generate_obs_structure(dataset)
        data += __memory_optimization_writing(
            dataset=dataset,
            obs_structure=obs_structure,
            prettyprint=prettyprint,
        )
    else:
        series_codes, obs_codes, group_codes = get_codes(
            dimension_code=dim,
            structure=dataset.structure,  # type: ignore[arg-type]
            data=dataset.data,
        )
        att_codes = [att.id for att in dataset.structure.components.attributes]
        series_att_codes = [x for x in series_codes if x in att_codes]
        obs_att_codes = [x for x in obs_codes if x in att_codes]

        series_codes = [x for x in series_codes if x not in series_att_codes]
        obs_codes = [x for x in obs_codes if x not in obs_att_codes]

        if group_codes:
            data += __group_processing(
                data=dataset.data,
                group_codes=group_codes,
                prettyprint=prettyprint,
            )

        data += __series_processing(
            data=dataset.data,
            series_codes=series_codes,
            series_att_codes=series_att_codes,
            obs_codes=obs_codes,
            obs_att_codes=obs_att_codes,
            prettyprint=prettyprint,
        )

    # Remove optional attributes empty data
    data = __remove_optional_attributes_empty_data(data)

    # Add to outfile
    outfile += data

    outfile += f"{child1}</{ABBR_MSG}:DataSet>"

    return outfile.replace("'", '"')


def __obs_processing(
    data: pd.DataFrame,
    obs_structure: Tuple[List[str], str, List[str]],
    prettyprint: bool = True,
) -> str:
    def __format_obs_str(element: Dict[str, Any]) -> str:
        child2 = "\t\t" if prettyprint else ""
        child3 = "\t\t\t" if prettyprint else ""
        child4 = "\t\t\t\t" if prettyprint else ""
        nl = "\n" if prettyprint else ""

        out = f"{child2}<{ABBR_GEN}:Obs>{nl}"
        # Obs Key writing
        out += f"{child3}<{ABBR_GEN}:ObsKey>{nl}"
        for k, v in element.items():
            if k in obs_structure[0]:
                out += f"{child4}{__value(k, v)}{nl}"
        out += f"{child3}</{ABBR_GEN}:ObsKey>{nl}"

        # Obs Value writing
        out += (
            f"{child3}<{ABBR_GEN}:ObsValue "
            f"value={str(element[obs_structure[1]])!r}/>{nl}"
        )

        if len(obs_structure[2]) > 0:
            # Obs Attributes writing
            out += f"{child3}<{ABBR_GEN}:Attributes>{nl}"
            for k, v in element.items():
                if k in obs_structure[2]:
                    out += f"{child4}{__value(k, v)}{nl}"
            out += f"{child3}</{ABBR_GEN}:Attributes>{nl}"

        out += f"{child2}</{ABBR_GEN}:Obs>{nl}"

        return out

    parser = lambda x: __format_obs_str(x)  # noqa: E731

    iterator = map(parser, data.to_dict(orient="records"))

    return "".join(iterator)


def __group_processing(
    data: pd.DataFrame,
    group_codes: List[Dict[str, Any]],
    prettyprint: bool = True,
) -> str:
    def __format_group_str(
        data_info: Dict[Any, Any],
        group_id: str,
        dimensions: List[str],
        attribute: str,
    ) -> str:
        """Formats a generic SDMX group using __value()."""
        child2 = "\t\t" if prettyprint else ""
        nl = "\n" if prettyprint else ""

        out_element = f'{child2}<{ABBR_GEN}:Group type="{group_id}">{nl}'

        # GroupKey block
        out_element += f"{child2}\t<{ABBR_GEN}:GroupKey>{nl}"
        for dim in dimensions:
            out_element += (
                f"{child2}\t\t{__value(dim, data_info.get(dim, ''))}{nl}"
            )
        out_element += f"{child2}\t</{ABBR_GEN}:GroupKey>{nl}"

        # Attributes block
        out_element += f"{child2}\t<{ABBR_GEN}:Attributes>{nl}"
        out_element += (
            f"{child2}\t\t{__value(attribute, data_info.get(attribute, ''))}"
            f"{nl}"
        )
        out_element += f"{child2}\t</{ABBR_GEN}:Attributes>{nl}"

        out_element += f"{child2}</{ABBR_GEN}:Group>{nl}"
        return out_element

    out_list: List[str] = []

    for group in group_codes:
        group_id = group["group_id"]
        dimensions = group["dimensions"]
        attribute = group["attribute"]
        group_keys = dimensions + [attribute]

        grouped_data = (
            data[group_keys]
            .drop_duplicates()
            .reset_index(drop=True)
            .to_dict(orient="records")
        )

        for record in grouped_data:
            out_list.append(
                __format_group_str(record, group_id, dimensions, attribute)
            )

    return "".join(out_list)


def __series_processing(
    data: pd.DataFrame,
    series_codes: List[str],
    series_att_codes: List[str],
    obs_codes: List[str],
    obs_att_codes: List[str],
    prettyprint: bool = True,
) -> str:
    def __generate_series_str() -> str:
        out_list: List[str] = []
        data.groupby(by=series_codes + series_att_codes)[data.columns].apply(
            lambda x: __format_dict_ser(out_list, x)
        )

        return "".join(out_list)

    def __format_dict_ser(
        output_list: List[str],
        group_data: Any,
    ) -> Any:
        obs_data = group_data[obs_codes + obs_att_codes].copy()
        data_dict["Series"][0]["Obs"] = obs_data.to_dict(orient="records")
        data_dict["Series"][0].update(
            {
                k: v
                for k, v in group_data[series_att_codes].iloc[0].items()
                if k in series_att_codes
            }
        )
        output_list.append(
            __format_ser_str(
                data_info=data_dict["Series"][0],
                series_codes=series_codes,
                series_att_codes=series_att_codes,
                obs_codes=obs_codes,
                obs_att_codes=obs_att_codes,
                prettyprint=prettyprint,
            )
        )
        del data_dict["Series"][0]

    # Getting each datapoint from data and creating dict
    data = data.sort_values(series_codes, axis=0)
    data_dict = {
        "Series": data[series_codes]
        .drop_duplicates()
        .reset_index(drop=True)
        .to_dict(orient="records")
    }

    out = __generate_series_str()

    return out


def __format_ser_str(
    data_info: Dict[Any, Any],
    series_codes: List[str],
    series_att_codes: List[str],
    obs_codes: List[str],
    obs_att_codes: List[str],
    prettyprint: bool,
) -> str:
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""
    child4 = "\t\t\t\t" if prettyprint else ""
    child5 = "\t\t\t\t\t" if prettyprint else ""
    nl = "\n" if prettyprint else ""

    out_element = f"{child2}<{ABBR_GEN}:Series>{nl}"

    # Series Key writing
    out_element += f"{child3}<{ABBR_GEN}:SeriesKey>{nl}"
    for k, v in data_info.items():
        if k in series_codes:
            out_element += f"{child4}{__value(k, v)}{nl}"
    out_element += f"{child3}</{ABBR_GEN}:SeriesKey>{nl}"

    # Series Attributes writing
    if len(series_att_codes) > 0:
        out_element += f"{child3}<{ABBR_GEN}:Attributes>{nl}"
        for k, v in data_info.items():
            if k in series_att_codes:
                out_element += f"{child4}{__value(k, v)}{nl}"
        out_element += f"{child3}</{ABBR_GEN}:Attributes>{nl}"

    # Obs writing
    for obs in data_info["Obs"]:
        out_element += f"{child3}<{ABBR_GEN}:Obs>{nl}"

        # Obs Dimension writing
        out_element += (
            f"{child4}<{ABBR_GEN}:ObsDimension "
            f"value={str(obs[obs_codes[0]])!r}/>{nl}"
        )
        # Obs Value writing
        out_element += (
            f"{child4}<{ABBR_GEN}:ObsValue value={str(obs_codes[1])!r}/>{nl}"
        )

        # Obs Attributes writing
        if len(obs_att_codes) > 0:
            out_element += f"{child4}<{ABBR_GEN}:Attributes>{nl}"
            for k, v in obs.items():
                if k in obs_att_codes:
                    out_element += f"{child5}{__value(k, v)}{nl}"
            out_element += f"{child4}</{ABBR_GEN}:Attributes>{nl}"
        out_element += f"{child3}</{ABBR_GEN}:Obs>{nl}"

    out_element += f"{child2}</{ABBR_GEN}:Series>{nl}"

    return out_element


def write(
    datasets: Sequence[PandasDataset],
    output_path: Optional[Union[str, Path]] = None,
    prettyprint: bool = True,
    header: Optional[Header] = None,
    dimension_at_observation: Optional[Dict[str, str]] = None,
) -> Optional[str]:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        datasets: The datasets to be written.
        output_path: The path to save the file.
        prettyprint: Prettyprint or not.
        header: The header to be used (generated if None).
        dimension_at_observation:
          The mapping between the dataset and the dimension at observation.

    Returns:
        The XML string if path is empty, None otherwise.
    """
    type_ = Format.DATA_SDMX_ML_2_1_GEN

    # Checking if we have datasets,
    # we need to ensure we can write them correctly
    check_content_dataset(datasets)
    content = {dataset.short_urn: dataset for dataset in datasets}

    if header is None:
        header = Header()

    # Checking the dimension at observation mapping
    dim_mapping = check_dimension_at_observation(
        datasets=content, dimension_at_observation=dimension_at_observation
    )
    header.structure = dim_mapping
    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, prettyprint=prettyprint)
    # Generating the header
    outfile += __write_header(header, prettyprint, data_message=True)
    # Writing the content
    outfile += __write_data_generic(content, dim_mapping, prettyprint)

    outfile += get_end_message(type_, prettyprint)

    output_path = (
        str(output_path) if isinstance(output_path, Path) else output_path
    )

    if output_path is None or output_path == "":
        return outfile

    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)

    return None



"""Collection of SDMX-JSON schemas for organisations."""

from collections import defaultdict
from typing import Dict, Sequence, Set

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import ItemSchemeType
from pysdmx.io.json.sdmxjson2.messages.pa import JsonProvisionAgreement
from pysdmx.model import DataflowRef, DataProvider, DataProviderScheme
from pysdmx.util import parse_item_urn, parse_urn


class JsonDataProviderScheme(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for a data provider scheme."""

    dataProviders: Sequence[DataProvider] = ()

    def __get_df_ref(self, ref: str) -> DataflowRef:
        a = parse_urn(ref)
        return DataflowRef(id=a.id, agency=a.agency, version=a.version)

    def to_model(
        self, pas: Sequence[JsonProvisionAgreement]
    ) -> DataProviderScheme:
        """Converts a JsonDataProviderScheme to a list of Organisations."""
        if pas:
            paprs: Dict[str, Set[DataflowRef]] = defaultdict(set)
            for pa in pas:
                df = self.__get_df_ref(pa.dataflow)
                ref = parse_item_urn(pa.dataProvider)
                paprs[f"{ref.agency}:{ref.item_id}"].add(df)
            provs = [
                DataProvider(
                    id=p.id,
                    name=p.name,
                    description=p.description,
                    contacts=p.contacts,
                    dataflows=list(paprs[f"{self.agency}:{p.id}"]),
                    annotations=[a.to_model() for a in self.annotations],
                )
                for p in self.dataProviders
            ]
        else:
            provs = [
                DataProvider(
                    id=p.id,
                    name=p.name,
                    description=p.description,
                    contacts=p.contacts,
                    annotations=[a.to_model() for a in self.annotations],
                )
                for p in self.dataProviders
            ]
        return DataProviderScheme(
            agency=self.agency,
            description=self.description,
            items=provs,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
        )


class JsonDataProviderSchemes(Struct, frozen=True):
    """SDMX-JSON payload for the list of data provider schemes."""

    dataProviderSchemes: Sequence[JsonDataProviderScheme]
    provisionAgreements: Sequence[JsonProvisionAgreement] = ()

    def to_model(self) -> Sequence[DataProviderScheme]:
        """Converts a JsonDataProviderSchemes to a list of Organisations."""
        return [
            s.to_model(self.provisionAgreements)
            for s in self.dataProviderSchemes
        ]


class JsonProviderMessage(Struct, frozen=True):
    """SDMX-JSON payload for /dataproviderscheme queries."""

    data: JsonDataProviderSchemes

    def to_model(self) -> Sequence[DataProviderScheme]:
        """Returns the requested list of data provider schemes."""
        return self.data.to_model()



"""Module for writing metadata to XML files."""

from pathlib import Path
from typing import Dict, Optional, Sequence, Union

from pysdmx.io.format import Format
from pysdmx.io.xml.__structure_aux_writer import (
    STR_DICT_TYPE_LIST_30,
    __write_structures,
)
from pysdmx.io.xml.__write_aux import (
    __write_header,
    create_namespaces,
    get_end_message,
)
from pysdmx.model.__base import MaintainableArtefact
from pysdmx.model.message import Header


def write(
    structures: Sequence[MaintainableArtefact],
    output_path: Optional[Union[str, Path]] = None,
    prettyprint: bool = True,
    header: Optional[Header] = None,
) -> Optional[str]:
    """This function writes a SDMX-ML file from the Message Content.

    Args:
        structures: The content to be written
        output_path: The path to save the file
        prettyprint: Prettyprint or not
        header: The header to be used (generated if None)

    Returns:
        The XML string if output_path is empty, None otherwise
    """
    type_ = Format.STRUCTURE_SDMX_ML_3_1
    elements = {structure.short_urn: structure for structure in structures}
    if header is None:
        header = Header()

    content: Dict[str, Dict[str, MaintainableArtefact]] = {}
    for urn, element in elements.items():
        list_ = STR_DICT_TYPE_LIST_30[type(element)]
        if list_ not in content:
            content[list_] = {}
        content[list_][urn] = element

    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, prettyprint=prettyprint)
    # Generating the header
    outfile += __write_header(header, prettyprint, data_message=False)
    # Writing the content
    outfile += __write_structures(content, prettyprint, references_30=True)

    outfile += get_end_message(type_, prettyprint)

    output_path = (
        str(output_path) if isinstance(output_path, Path) else output_path
    )

    if output_path is None or output_path == "":
        return outfile

    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)
    return None



"""Collection of GDS-JSON schemas for GDS urn_resolver."""

from typing import List

from msgspec import Struct

from pysdmx.model.gds import GdsUrnResolver, ResolverResult


class JsonUrnResolverResult(Struct, frozen=True):
    """GDS-JSON payload for a single URN resolver result."""

    api_version: str
    query: str
    query_response_status_code: int

    def to_model(self) -> ResolverResult:
        """Returns a ResolverResult object."""
        return ResolverResult(
            api_version=self.api_version,
            query=self.query,
            status_code=self.query_response_status_code,
        )


class JsonUrnResolverMessage(Struct, frozen=True):
    """GDS-JSON payload for /urn_resolver queries."""

    agency_id: str
    resource_id: str
    version: str
    sdmx_type: str
    resolver_results: List[JsonUrnResolverResult]

    def to_model(self) -> GdsUrnResolver:
        """Returns a GdsUrnResolver object."""
        return GdsUrnResolver(
            agency=self.agency_id,
            resource_id=self.resource_id,
            version=self.version,
            sdmx_type=self.sdmx_type,
            resolver_results=[a.to_model() for a in self.resolver_results],
        )



"""Processes the input that comes into read_sdmx function."""

import csv
import os.path
from io import BytesIO, StringIO, TextIOWrapper
from json import JSONDecodeError, loads
from os import PathLike
from pathlib import Path
from typing import Optional, Tuple, Union

from httpx import Client as httpx_Client
from httpx import HTTPStatusError, create_ssl_context

from pysdmx.errors import Invalid, NotImplemented
from pysdmx.io.format import Format
from pysdmx.util import map_httpx_errors

SCHEMA_ROOT_31 = "http://www.sdmx.org/resources/sdmxml/schemas/v3_1/"
SCHEMA_ROOT_30 = "http://www.sdmx.org/resources/sdmxml/schemas/v3_0/"


def __remove_bom(input_string: str) -> str:
    return input_string.replace("\ufeff", "")


def __check_xml(input_str: str) -> bool:
    return input_str[:5] == "<?xml"


def __check_csv(input_str: str) -> bool:
    try:
        max_length = len(input_str) if len(input_str) < 2048 else 2048
        dialect = csv.Sniffer().sniff(input_str[:max_length])
        control_csv_format = (
            dialect.delimiter == "," and dialect.quotechar == '"'
        )
        # Check we can access the data and it is not empty
        if (
            len(input_str.splitlines()) > 1
            or input_str.splitlines()[0].count(",") > 1
        ) and control_csv_format:
            return True
    except Exception:
        return False
    return False


def __check_json(input_str: str) -> bool:
    try:
        loads(input_str)
        return True
    except JSONDecodeError:
        return False


def __get_sdmx_ml_flavour(input_str: str) -> Tuple[str, Format]:
    flavour_check = input_str[:1000].lower()
    if ":generic" in flavour_check:
        return input_str, Format.DATA_SDMX_ML_2_1_GEN

    if ":structurespecificdata" in flavour_check:
        if SCHEMA_ROOT_30 in flavour_check:
            return input_str, Format.DATA_SDMX_ML_3_0
        elif SCHEMA_ROOT_31 in flavour_check:
            return input_str, Format.DATA_SDMX_ML_3_1
        else:
            return input_str, Format.DATA_SDMX_ML_2_1_STR
    if ":structure" in flavour_check:
        if SCHEMA_ROOT_30 in flavour_check:
            return input_str, Format.STRUCTURE_SDMX_ML_3_0
        elif SCHEMA_ROOT_31 in flavour_check:
            return input_str, Format.STRUCTURE_SDMX_ML_3_1
        else:
            return input_str, Format.STRUCTURE_SDMX_ML_2_1
    if ":registryinterface" in flavour_check:
        return input_str, Format.REGISTRY_SDMX_ML_2_1
    if ":error" in flavour_check:
        return input_str, Format.ERROR_SDMX_ML_2_1
    raise Invalid("Validation Error", "Cannot parse input as SDMX-ML.")


def __get_sdmx_csv_flavour(input_str: str) -> Tuple[str, Format]:
    headers = csv.reader(StringIO(input_str)).__next__()
    if "DATAFLOW" in headers:
        return input_str, Format.DATA_SDMX_CSV_1_0_0
    elif "STRUCTURE" in headers and "STRUCTURE_ID" in headers:
        return input_str, Format.DATA_SDMX_CSV_2_0_0
    raise Invalid("Validation Error", "Cannot parse input as SDMX-CSV.")


def __check_sdmx_str(input_str: str) -> Tuple[str, Format]:
    """Attempts to infer the SDMX format of the input string."""
    if __check_xml(input_str):
        return __get_sdmx_ml_flavour(input_str)
    if __check_csv(input_str):
        return __get_sdmx_csv_flavour(input_str)
    if __check_json(input_str):
        raise NotImplemented("JSON formats reading are not supported yet")
    raise Invalid("Validation Error", "Cannot parse input as SDMX.")


def process_string_to_read(  # noqa: C901
    sdmx_document: Union[str, Path, BytesIO],
    pem: Optional[Union[str, Path]] = None,
) -> Tuple[str, Format]:
    """Processes the input that comes into read_sdmx function.

    Automatically detects the format of the input. The input can be a file,
    URL, or string.

    Args:
        sdmx_document: Path to file, URL, or string.
        pem: Path to a PEM file for SSL verification when reading from a URL.

    Returns:
        tuple: Tuple containing the parsed input and the format of the input.

    Raises:
        Invalid: If the input cannot be parsed as SDMX.
    """
    if isinstance(sdmx_document, str) and os.path.exists(sdmx_document):
        sdmx_document = Path(sdmx_document)
    # Read file as string
    if isinstance(sdmx_document, (Path, PathLike)):
        with open(
            sdmx_document, "r", encoding="utf-8-sig", errors="replace"
        ) as f:
            out_str = f.read()

    # Read from BytesIO
    elif isinstance(sdmx_document, BytesIO):
        text_wrap = TextIOWrapper(
            sdmx_document, encoding="utf-8", errors="replace"
        )
        out_str = text_wrap.read()

    elif isinstance(sdmx_document, str):
        if sdmx_document.startswith("http"):
            if pem is not None and isinstance(pem, Path):
                pem = str(pem)
            if pem is not None and pem and not os.path.exists(pem):
                raise Invalid(
                    "Validation Error",
                    f"PEM file {pem} does not exist.",
                )
            # Read from URL
            ssl_context = (
                create_ssl_context(
                    verify=pem,
                )
                if pem
                else create_ssl_context()
            )
            try:
                out_str = ""
                with httpx_Client(verify=ssl_context) as client:
                    response = client.get(sdmx_document, timeout=60)
                    try:
                        response.raise_for_status()
                    except HTTPStatusError as e:
                        if "<?xml" in e.response.text:
                            out_str = e.response.text
                        else:
                            map_httpx_errors(e)
                    else:
                        out_str = response.text
            except Exception as e:
                raise Invalid(
                    "Validation Error",
                    f"Cannot retrieve a SDMX Message "
                    f"from URL: {sdmx_document}. Error message: {e}",
                ) from None
        else:
            out_str = sdmx_document
    else:
        raise Invalid(
            "Validation Error",
            f"Cannot parse input of type {type(sdmx_document)}.",
        )

    out_str = __remove_bom(out_str)

    return __check_sdmx_str(out_str)



"""Collection of SDMX-JSON schemas for generic structure messages."""

from typing import Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.agency import JsonAgencyScheme
from pysdmx.io.json.sdmxjson2.messages.category import (
    JsonCategorisation,
    JsonCategoryScheme,
)
from pysdmx.io.json.sdmxjson2.messages.code import (
    JsonCodelist,
    JsonHierarchy,
    JsonHierarchyAssociation,
    JsonValuelist,
)
from pysdmx.io.json.sdmxjson2.messages.concept import JsonConceptScheme
from pysdmx.io.json.sdmxjson2.messages.core import JsonHeader
from pysdmx.io.json.sdmxjson2.messages.dataflow import JsonDataflow
from pysdmx.io.json.sdmxjson2.messages.dsd import JsonDataStructure
from pysdmx.io.json.sdmxjson2.messages.map import (
    JsonRepresentationMap,
    JsonStructureMap,
)
from pysdmx.io.json.sdmxjson2.messages.pa import JsonProvisionAgreement
from pysdmx.io.json.sdmxjson2.messages.provider import JsonDataProviderScheme
from pysdmx.io.json.sdmxjson2.messages.vtl import (
    JsonCustomTypeScheme,
    JsonNamePersonalisationScheme,
    JsonRulesetScheme,
    JsonTransformationScheme,
    JsonUserDefinedOperatorScheme,
    JsonVtlMappingScheme,
)
from pysdmx.model.__base import MaintainableArtefact
from pysdmx.model.message import StructureMessage


class JsonStructures(Struct, frozen=True):
    """The allowed strutures."""

    dataStructures: Sequence[JsonDataStructure] = ()
    categorySchemes: Sequence[JsonCategoryScheme] = ()
    conceptSchemes: Sequence[JsonConceptScheme] = ()
    codelists: Sequence[JsonCodelist] = ()
    valueLists: Sequence[JsonValuelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()
    hierarchyAssociations: Sequence[JsonHierarchyAssociation] = ()
    agencySchemes: Sequence[JsonAgencyScheme] = ()
    dataProviderSchemes: Sequence[JsonDataProviderScheme] = ()
    dataflows: Sequence[JsonDataflow] = ()
    provisionAgreements: Sequence[JsonProvisionAgreement] = ()
    structureMaps: Sequence[JsonStructureMap] = ()
    representationMaps: Sequence[JsonRepresentationMap] = ()
    categorisations: Sequence[JsonCategorisation] = ()
    customTypeSchemes: Sequence[JsonCustomTypeScheme] = ()
    vtlMappingSchemes: Sequence[JsonVtlMappingScheme] = ()
    namePersonalisationSchemes: Sequence[JsonNamePersonalisationScheme] = ()
    rulesetSchemes: Sequence[JsonRulesetScheme] = ()
    transformationSchemes: Sequence[JsonTransformationScheme] = ()
    userDefinedOperatorSchemes: Sequence[JsonUserDefinedOperatorScheme] = ()

    def to_model(self) -> Sequence[MaintainableArtefact]:
        """Map to pysdmx artefacts."""
        structures = []  # type: ignore[var-annotated]
        structures.extend(
            i.to_model(
                self.conceptSchemes, self.codelists, self.valueLists, ()
            )
            for i in self.dataStructures
        )
        structures.extend(i.to_model() for i in self.categorySchemes)
        structures.extend(
            i.to_model(self.codelists) for i in self.conceptSchemes
        )
        structures.extend(i.to_model() for i in self.codelists)
        structures.extend(i.to_model() for i in self.valueLists)
        structures.extend(i.to_model(self.codelists) for i in self.hierarchies)
        structures.extend(
            i.to_model(self.hierarchies, self.codelists)
            for i in self.hierarchyAssociations
        )
        structures.extend(
            i.to_model(self.dataflows) for i in self.agencySchemes
        )
        structures.extend(
            i.to_model(self.provisionAgreements)
            for i in self.dataProviderSchemes
        )
        structures.extend(
            i.to_model(
                self.dataStructures,
                self.conceptSchemes,
                self.valueLists,
                self.codelists,
            )
            for i in self.dataflows
        )
        structures.extend(i.to_model() for i in self.provisionAgreements)
        structures.extend(
            i.to_model(self.representationMaps) for i in self.structureMaps
        )
        structures.extend(i.to_model() for i in self.categorisations)
        structures.extend(i.to_model() for i in self.customTypeSchemes)
        structures.extend(i.to_model() for i in self.vtlMappingSchemes)
        structures.extend(
            i.to_model() for i in self.namePersonalisationSchemes
        )
        structures.extend(i.to_model() for i in self.rulesetSchemes)
        structures.extend(
            i.to_model(
                self.customTypeSchemes,
                self.vtlMappingSchemes,
                self.namePersonalisationSchemes,
                self.rulesetSchemes,
                self.userDefinedOperatorSchemes,
            )
            for i in self.transformationSchemes
        )
        structures.extend(
            i.to_model() for i in self.userDefinedOperatorSchemes
        )
        for rm in self.representationMaps:
            multi = bool(len(rm.source) > 1 or len(rm.target) > 1)
            structures.append(rm.to_model(multi))
        return structures


class JsonStructureMessage(Struct, frozen=True):
    """A generic SDMX-JSON 2.0 Structure message."""

    meta: JsonHeader
    data: JsonStructures

    def to_model(self) -> StructureMessage:
        """Map to pysdmx message class."""
        header = self.meta.to_model()
        structures = self.data.to_model()
        return StructureMessage(header, structures)



"""Collection of SDMX-JSON schemas for provision agreements."""

from typing import Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import (
    MaintainableType,
)
from pysdmx.model import ProvisionAgreement


class JsonProvisionAgreement(MaintainableType, frozen=True):
    """SDMX-JSON payload for a provision agreement."""

    dataflow: str = ""
    dataProvider: str = ""

    def to_model(self) -> ProvisionAgreement:
        """Converts a FusionPA to a standard provision agreement."""
        return ProvisionAgreement(
            id=self.id,
            agency=self.agency,
            name=self.name,
            description=self.description,
            version=self.version,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            dataflow=self.dataflow,
            provider=self.dataProvider,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
        )


class JsonProvisionAgreements(Struct, frozen=True):
    """SDMX-JSON payload for provision agreements."""

    provisionAgreements: Sequence[JsonProvisionAgreement]

    def to_model(self) -> Sequence[ProvisionAgreement]:
        """Returns the requested provision agreements."""
        return [pa.to_model() for pa in self.provisionAgreements]


class JsonProvisionAgreementsMessage(Struct, frozen=True):
    """SDMX-JSON payload for /provisionagreement queries."""

    data: JsonProvisionAgreements

    def to_model(self) -> Sequence[ProvisionAgreement]:
        """Returns the requested provision agreements."""
        return self.data.to_model()



"""Build SDMX-REST reference metadata queries."""

from datetime import datetime
from enum import Enum
from typing import Optional, Sequence, Union

import msgspec
from msgspec.json import Decoder

from pysdmx.api.qb.structure import _API_RESOURCES, StructureType
from pysdmx.api.qb.util import REST_ALL, REST_LATEST, ApiVersion, CoreQuery
from pysdmx.errors import Invalid
from pysdmx.io.format import RefMetaFormat
from pysdmx.model import Reference
from pysdmx.util import parse_urn


class RefMetaDetail(Enum):
    """The desired amount of information to be returned."""

    FULL = "full"
    ALL_STUBS = "allstubs"


class _RefMetaCoreQuery(CoreQuery, frozen=True, omit_defaults=True):
    def _check_version(self, version: ApiVersion) -> None:
        if version < ApiVersion.V2_0_0:
            raise Invalid(
                "Invalid Request",
                (
                    "Queries for reference metadata are not supported"
                    f"in SDMX-REST {version.value}."
                ),
            )

    def _check_as_of(
        self, as_of: Optional[datetime], version: ApiVersion
    ) -> None:
        if as_of and version < ApiVersion.V2_2_0:
            raise Invalid(
                "Validation Error",
                f"as_of not supported in {version.value}.",
            )

    def _get_as_of_value(self, as_of: Optional[datetime]) -> str:
        return f'&asOf={as_of.isoformat("T", "seconds")}' if as_of else ""

    def _get_short_qs(
        self, detail: RefMetaDetail, as_of: Optional[datetime]
    ) -> str:
        qs = ""
        if detail != RefMetaDetail.FULL:
            qs = super()._append_qs_param(qs, detail.value, "detail")
        qs = super()._append_qs_param(
            qs,
            as_of,
            "asOf",
            as_of.isoformat("T", "seconds") if as_of else None,
        )
        if qs:
            qs = f"?{qs}"
        return qs


class RefMetaByMetadatasetQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata with metadataset identification details.

    Attributes:
        provider_id: The id(s) of the data provider.
        metadataset_id: The id(s) of the metadataset(s) to be returned.
        version: The version(s) of the metadataset(s) to be returned.
        detail: The desired amount of information to be returned.
        as_of: Retrieve the metadata as they were at the specified point
            in time (aka time travel).
    """

    provider_id: Union[str, Sequence[str]] = REST_ALL
    metadataset_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    detail: RefMetaDetail = RefMetaDetail.FULL
    as_of: Optional[datetime] = None

    @staticmethod
    def from_ref(ref: Union[Reference, str]) -> "RefMetaByMetadatasetQuery":
        """Create a RefMetaByMetadatasetQuery out of the supplied reference.

        Args:
            ref: Either a reference object or an SDMX urn.

        Returns:
            A RefMetaByMetadatasetQuery to retrieve the supplied reference.

        Raises:
            Invalid: If reference is a string and it is not an SDMX urn,
                or if the artefact type is not MetadataSet.
        """
        if isinstance(ref, str):
            ref = parse_urn(ref)  # type: ignore[assignment]
        if ref.sdmx_type != "MetadataSet":  # type: ignore[union-attr]
            raise Invalid(
                "Unexpected artefact type",
                (
                    "Only references of type MetadataSet can be converted"
                    "into a RefMetaByMetadatasetQuery"
                ),
                {"received_type": ref.sdmx_type},  # type: ignore[union-attr]
            )
        return RefMetaByMetadatasetQuery(
            ref.agency,  # type: ignore[union-attr]
            ref.id,  # type: ignore[union-attr]
            ref.version,  # type: ignore[union-attr]
        )

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        super()._check_as_of(self.as_of, version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_mds_decoder

    def _create_full_query(self, ver: ApiVersion) -> str:
        p = super()._join_mult(self.provider_id)
        i = super()._join_mult(self.metadataset_id)
        v = super()._join_mult(self.version)
        ao = super()._get_as_of_value(self.as_of)
        return (
            f"/metadata/metadataset/{p}/{i}/{v}?detail={self.detail.value}{ao}"
        )

    def _create_short_query(self, ver: ApiVersion) -> str:
        v = f"/{self.version}" if self.version != REST_LATEST else ""
        i = (
            f"/{self.metadataset_id}{v}"
            if v or self.metadataset_id != REST_ALL
            else ""
        )
        p = (
            f"/{self.provider_id}{i}"
            if i or self.provider_id != REST_ALL
            else ""
        )
        q = super()._get_short_qs(self.detail, self.as_of)
        return f"/metadata/metadataset{p}{q}"


class RefMetaByStructureQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata reported against one or more structures.

    Attributes:
        artefact_type: The type of structural metadata to which the
            reference metadata to be returned are attached.
        agency_id: The agency (or agencies) maintaining the artefact(s)
            to which the reference metadata to be returned are attached.
        resource_id: The id(s) of the artefact(s) to which the reference
            metadata to be returned are attached.
        version: The version(s) of the artefact(s) to which the reference
            metadata to be returned are attached.
        detail: The desired amount of information to be returned.
        as_of: Retrieve the metadata as they were at the specified point
            in time (aka time travel).
    """

    artefact_type: StructureType = StructureType.ALL
    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    detail: RefMetaDetail = RefMetaDetail.FULL
    as_of: Optional[datetime] = None

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        super()._check_as_of(self.as_of, version)
        self.__check_artefact_type(self.artefact_type, version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_struct_decoder

    def __check_artefact_type(
        self, atyp: StructureType, version: ApiVersion
    ) -> None:
        if atyp not in _API_RESOURCES[version.name.replace("_", ".")]:
            raise Invalid(
                "Validation Error",
                f"{atyp} is not valid for SDMX-REST {version.name}.",
            )

    def _create_full_query(self, ver: ApiVersion) -> str:
        a = super()._join_mult(self.agency_id)
        r = super()._join_mult(self.resource_id)
        v = super()._join_mult(self.version)
        ao = super()._get_as_of_value(self.as_of)
        return (
            f"/metadata/structure/{self.artefact_type.value}/{a}/{r}/{v}"
            f"?detail={self.detail.value}{ao}"
        )

    def _create_short_query(self, ver: ApiVersion) -> str:
        v = f"/{self.version}" if self.version != REST_LATEST else ""
        r = (
            f"/{self.resource_id}{v}"
            if v or self.resource_id != REST_ALL
            else ""
        )
        a = f"/{self.agency_id}{r}" if r or self.agency_id != REST_ALL else ""
        t = (
            f"/{self.artefact_type.value}{a}"
            if a or self.artefact_type != StructureType.ALL
            else ""
        )
        q = super()._get_short_qs(self.detail, self.as_of)
        return f"/metadata/structure{t}{q}"


class RefMetaByMetadataflowQuery(
    _RefMetaCoreQuery,
    frozen=True,
    omit_defaults=True,
):
    """A query for reference metadata reported for metadataflows.

    Attributes:
        agency_id: The agency (or agencies) maintaining the metadataflow(s)
            of which reference metadata need to be returned.
        resource_id: The id(s) of the metadataflow(s) of which reference
            metadata need to be returned.
        version: The version(s) of the metadataflows(s) of which reference
            metadata need to be returned.
        provider_id: The id(s) of the providers that provided the reference
            metadata to be returned.
        detail: The desired amount of information to be returned.
        as_of: Retrieve the metadata as they were at the specified point
            in time (aka time travel).
    """

    agency_id: Union[str, Sequence[str]] = REST_ALL
    resource_id: Union[str, Sequence[str]] = REST_ALL
    version: Union[str, Sequence[str]] = REST_LATEST
    provider_id: Union[str, Sequence[str]] = REST_ALL
    detail: RefMetaDetail = RefMetaDetail.FULL
    as_of: Optional[datetime] = None

    def _validate_query(self, version: ApiVersion) -> None:
        super().validate()
        super()._check_version(version)
        super()._check_as_of(self.as_of, version)

    def _get_decoder(self) -> Decoder:  # type: ignore[type-arg]
        return _by_flow_decoder

    def _create_full_query(self, ver: ApiVersion) -> str:
        a = super()._join_mult(self.agency_id)
        r = super()._join_mult(self.resource_id)
        v = super()._join_mult(self.version)
        p = super()._join_mult(self.provider_id)
        ao = super()._get_as_of_value(self.as_of)
        return (
            f"/metadata/metadataflow/{a}/{r}/{v}/{p}"
            f"?detail={self.detail.value}{ao}"
        )

    def _create_short_query(self, ver: ApiVersion) -> str:
        p = f"/{self.provider_id}" if self.provider_id != REST_ALL else ""
        v = f"/{self.version}{p}" if p or self.version != REST_LATEST else ""
        r = (
            f"/{self.resource_id}{v}"
            if v or self.resource_id != REST_ALL
            else ""
        )
        a = f"/{self.agency_id}{r}" if r or self.agency_id != REST_ALL else ""
        q = super()._get_short_qs(self.detail, self.as_of)
        return f"/metadata/metadataflow{a}{q}"


_by_mds_decoder = msgspec.json.Decoder(RefMetaByMetadatasetQuery)
_by_struct_decoder = msgspec.json.Decoder(RefMetaByStructureQuery)
_by_flow_decoder = msgspec.json.Decoder(RefMetaByMetadataflowQuery)


__all__ = [
    "RefMetaDetail",
    "RefMetaFormat",
    "RefMetaByMetadataflowQuery",
    "RefMetaByMetadatasetQuery",
    "RefMetaByStructureQuery",
]



"""Model for SDMX Reference Metadata.

Reference metadata are quite generic (and therefore powerful) in SDMX.
Though they are typically used to provide additional information about
statistical data, such as information about the general methodological
and quality aspects of the statistical production process, they can also
be used to drive process steps such as validation or mapping, for
example by providing configuration details in a metadata report.
"""

from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional, Sequence

from msgspec import Struct

from pysdmx.model.__base import Annotation, MaintainableArtefact
from pysdmx.model.concept import Facets
from pysdmx.model.dataset import ActionType


class MetadataAttribute(
    Struct, frozen=True, omit_defaults=True, repr_omit_defaults=True
):
    """An entry in a metadata report.

    An attribute is iterable, as it may contain other attributes.

    Attributes:
        id: The identifier of the attribute (e.g. "License").
        value: The attribute value (e.g. "BSD-2-Clause").
        attributes: The list of "children" attributes (i.e. attributes
            can be nested).
    """

    id: str
    value: Optional[Any] = None
    attributes: Sequence["MetadataAttribute"] = ()
    annotations: Sequence[Annotation] = ()
    format: Optional[Facets] = None

    def __iter__(self) -> Iterator["MetadataAttribute"]:
        """Return an iterator over the list of attributes."""
        yield from self.attributes

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # str is taken as a Sequence, so we need to check it's not a str
            if isinstance(value, Sequence) and not isinstance(value, str):
                # Handle non-empty lists
                if not value:
                    continue
                class_name = value[0].__class__.__name__
                value = f"{len(value)} {class_name.lower()}s"

            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # Omit empty sequences
            if isinstance(value, (list, tuple, set)) and not value:
                continue
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"


class MetadataReport(MaintainableArtefact, frozen=True, omit_defaults=True):
    """An organized collection of metadata.

    A metadata report is iterable and it is also possible to directly
    retrieve an attribute using its ID.

    Attributes:
        metadataflow: The URN of the dataflow for which the report
            belongs.
        targets: The URN(s) of SDMX artefact(s) to which the report relates.
        attributes: The list of metadata attributes included in the report.
            Attributes may contain other attributes.
        metadataProvisionAgreement: reference to a metadata provision
            agreement
        publicationPeriod: The reporting period to which the metadata report
            relates
        publicationYear: The year when the report was published
        reportingBegin: The oldest period to which the report relates.
        reportingEnd: The most recent period to which the report relates.
        action: The action to be performed by the receiver.
    """

    metadataflow: str = ""
    targets: Sequence[str] = ()
    attributes: Sequence[MetadataAttribute] = ()
    metadataProvisionAgreement: Optional[str] = None
    publicationPeriod: Optional[str] = None
    publicationYear: Optional[str] = None
    reportingBegin: Optional[str] = None
    reportingEnd: Optional[str] = None
    action: Optional[ActionType] = None

    def __iter__(self) -> Iterator[MetadataAttribute]:
        """Return an iterator over the list of report attributes."""
        yield from self.attributes

    def __len__(self) -> int:
        """Return the number of attributes in the report."""
        return self.__get_count(self.attributes)

    def __getitem__(self, id_: str) -> Optional[MetadataAttribute]:
        """Return the attribute identified by the given ID."""
        return self.__extract_attr(self.attributes, id_)

    def __get_count(self, attributes: Sequence[MetadataAttribute]) -> int:
        """Return the number of attributes at all levels."""
        count = len(attributes)
        for attr in attributes:
            if attr.attributes:
                count += self.__get_count(attr.attributes)
        return count

    def __extract_attr(
        self, attributes: Sequence[MetadataAttribute], id_: str
    ) -> Optional[MetadataAttribute]:
        if "." in id_:
            ids = id_.split(".")
            out = list(filter(lambda attr: attr.id == ids[0], attributes))
            if out:
                pkey = ".".join(ids[1:])
                return self.__extract_attr(out[0].attributes, pkey)
        else:
            out = list(filter(lambda attr: attr.id == id_, attributes))
            if out:
                return out[0]
        return None

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # str is taken as a Sequence, so we need to check it's not a str
            if isinstance(value, Sequence) and not isinstance(value, str):
                # Handle non-empty lists
                if not value:
                    continue
                class_name = value[0].__class__.__name__
                if class_name == "MetadataAttribute":
                    class_name = "Metadata Attribute"
                value = f"{len(value)} {class_name.lower()}s"

            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # Omit empty sequences
            if isinstance(value, (list, tuple, set)) and not value:
                continue
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"


def merge_attributes(
    attrs: Sequence[MetadataAttribute],
) -> Sequence[MetadataAttribute]:
    """Groups together the values of attributes with the same ID.

    The function assumes that an attribute will either have a
    value or will act as a container for other attributes. In
    case the attribute contains other attributes AND has a value,
    this function will NOT work as expected.

    Args:
        attrs: The list of attributes to be merged

    Returns:
        The list of (possibly merged) attributes
    """
    by_id: Dict[str, List[Any]] = defaultdict(list)
    sub_id = []

    for attr in attrs:
        if attr.attributes:
            sub_id.append(
                MetadataAttribute(
                    attr.id,
                    attr.value,
                    merge_attributes(attr.attributes),
                )
            )
        else:
            by_id[attr.id].append(attr.value)

    out = []
    out.extend(sub_id)
    for k, v in by_id.items():
        val = v if len(v) > 1 else v[0]
        out.append(MetadataAttribute(k, val))
    return out



"""Collection of SDMX-JSON schemas for reference metadata queries."""

from typing import Any, Optional, Sequence

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import (
    IdentifiableType,
    ItemSchemeType,
    JsonHeader,
    JsonTextFormat,
    get_facets,
)
from pysdmx.model.dataset import ActionType
from pysdmx.model.message import MetadataMessage
from pysdmx.model.metadata import (
    MetadataAttribute,
    MetadataReport,
    merge_attributes,
)


class JsonMetadataAttribute(IdentifiableType, frozen=True, omit_defaults=True):
    """SDMX-JSON payload for a metadata attribute."""

    value: Optional[Any] = None
    attributes: Sequence["JsonMetadataAttribute"] = ()
    format: Optional[JsonTextFormat] = None

    def to_model(self) -> MetadataAttribute:
        """Converts a JsonMetadataAttribute to a standard attribute."""
        attrs = [a.to_model() for a in self.attributes]
        attrs = merge_attributes(attrs)  # type: ignore[assignment]
        return MetadataAttribute(
            id=self.id,
            value=self.value,
            attributes=attrs,
            annotations=[a.to_model() for a in self.annotations],
            format=get_facets(self.format) if self.format else None,
        )


class JsonMetadataReport(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for a metadata report."""

    metadataflow: str = ""
    targets: Sequence[str] = ()
    attributes: Sequence[JsonMetadataAttribute] = ()
    metadataProvisionAgreement: Optional[str] = None
    publicationPeriod: Optional[str] = None
    publicationYear: Optional[str] = None
    reportingBegin: Optional[str] = None
    reportingEnd: Optional[str] = None
    action: Optional[str] = None

    def to_model(self) -> MetadataReport:
        """Converts a JsonMetadataReport to a standard report."""
        attrs = [a.to_model() for a in self.attributes]
        attrs = merge_attributes(attrs)  # type: ignore[assignment]
        return MetadataReport(
            annotations=[a.to_model() for a in self.annotations],
            id=self.id,
            name=self.name,
            description=self.description,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            version=self.version,
            agency=self.agency,
            is_external_reference=self.isExternalReference,
            metadataflow=self.metadataflow,
            targets=self.targets,
            attributes=attrs,
            metadataProvisionAgreement=self.metadataProvisionAgreement,
            publicationPeriod=self.publicationPeriod,
            publicationYear=self.publicationYear,
            reportingBegin=self.reportingBegin,
            reportingEnd=self.reportingEnd,
            action=ActionType(self.action) if self.action else None,
        )


class JsonMetadataSets(Struct, frozen=True):
    """SDMX-JSON payload for the list of metadata sets."""

    metadataSets: Sequence[JsonMetadataReport]

    def to_model(self) -> Sequence[MetadataReport]:
        """Returns the requested metadata report(s)."""
        return [r.to_model() for r in self.metadataSets]


class JsonMetadataMessage(Struct, frozen=True):
    """SDMX-JSON payload for /metadata queries."""

    meta: JsonHeader
    data: JsonMetadataSets

    def to_model(self) -> MetadataMessage:
        """Returns the requested metadata report(s)."""
        header = self.meta.to_model()
        reports = self.data.to_model()
        return MetadataMessage(header, reports)



"""Collection of SDMX-JSON schemas for codes and codelists."""

from datetime import datetime
from datetime import timezone as tz
from typing import Optional, Sequence, Tuple

from msgspec import Struct

from pysdmx.io.json.sdmxjson2.messages.core import (
    ItemSchemeType,
    JsonAnnotation,
    JsonLink,
    MaintainableType,
    NameableType,
)
from pysdmx.model import (
    Code,
    Codelist,
    HierarchicalCode,
    Hierarchy,
    HierarchyAssociation,
)
from pysdmx.util import find_by_urn, parse_item_urn


class JsonCode(NameableType, frozen=True):
    """SDMX-JSON payload for codes."""

    parent: Optional[str] = None

    def __handle_date(self, datestr: str) -> datetime:
        return datetime.strptime(datestr, "%Y-%m-%dT%H:%M:%S%z")

    def __get_val(
        self, a: JsonAnnotation
    ) -> Tuple[Optional[datetime], Optional[datetime]]:
        vals = a.title.split("/")  # type: ignore[union-attr]
        if a.title.startswith("/"):  # type: ignore[union-attr]
            return (None, self.__handle_date(vals[1]))
        else:
            valid_from = self.__handle_date(vals[0])
            valid_to = self.__handle_date(vals[1]) if vals[1] else None
            return (valid_from, valid_to)

    def to_model(self) -> Code:
        """Converts a JsonCode to a standard code."""
        if self.annotations:
            vp = [
                a for a in self.annotations if a.type == "FR_VALIDITY_PERIOD"
            ]
        else:
            vp = None
        vf, vt = self.__get_val(vp[0]) if vp else (None, None)
        return Code(
            id=self.id,
            name=self.name,
            description=self.description,
            valid_from=vf,
            valid_to=vt,
            annotations=[a.to_model() for a in self.annotations],
        )


class JsonCodelist(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for a codelist."""

    codes: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonCodelist to a standard codelist."""
        return Codelist(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[i.to_model() for i in self.codes],
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
        )


class JsonValuelist(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for a valuelist."""

    valueItems: Sequence[JsonCode] = ()

    def to_model(self) -> Codelist:
        """Converts a JsonValuelist to a standard codelist."""
        return Codelist(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            items=[i.to_model() for i in self.valueItems],
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            sdmx_type="valuelist",
        )


class JsonCodelists(Struct, frozen=True):
    """SDMX-JSON payload for lists of codes."""

    codelists: Sequence[JsonCodelist] = ()
    valuelists: Sequence[JsonValuelist] = ()


class JsonCodelistMessage(Struct, frozen=True):
    """SDMX-JSON payload for /codelist queries."""

    data: JsonCodelists

    def to_model(self) -> Codelist:
        """Returns the requested codelist."""
        if self.data.codelists:
            return self.data.codelists[0].to_model()
        else:
            return self.data.valuelists[0].to_model()


class JsonHierarchicalCode(Struct, frozen=True):
    """Fusion-JSON payload for hierarchical codes."""

    id: str
    code: str
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    annotations: Optional[Sequence[JsonAnnotation]] = None
    hierarchicalCodes: Sequence["JsonHierarchicalCode"] = ()

    def __find_code(self, codelists: Sequence[Codelist], urn: str) -> Code:
        r = parse_item_urn(urn)
        f = [
            c
            for c in codelists
            if (
                c.agency == r.agency
                and c.id == r.id
                and c.version == r.version
            )
        ]
        return [c for c in f[0].codes if c.id == r.item_id][0]

    def to_model(self, codelists: Sequence[Codelist]) -> HierarchicalCode:
        """Converts a JsonHierarchicalCode to a hierachical code."""
        if codelists:
            code = self.__find_code(codelists, self.code)
            name = code.name
            description = code.description
        else:
            r = parse_item_urn(self.code)
            code = Code(r.item_id)
            name = None
            description = None
        codes = [c.to_model(codelists) for c in self.hierarchicalCodes]
        vf = self.validFrom.replace(tzinfo=tz.utc) if self.validFrom else None
        vt = self.validTo.replace(tzinfo=tz.utc) if self.validTo else None
        return HierarchicalCode(
            code.id,
            name,
            description,
            code.valid_from,
            code.valid_to,
            vf,
            vt,
            codes,
        )


class JsonHierarchy(ItemSchemeType, frozen=True):
    """SDMX-JSON payload for a hierarchy."""

    hierarchicalCodes: Sequence[JsonHierarchicalCode] = ()

    def to_model(self, codelists: Sequence[JsonCodelist]) -> Hierarchy:
        """Converts a JsonHierarchy to a standard hierarchy."""
        cls = [cl.to_model() for cl in codelists]
        return Hierarchy(
            id=self.id,
            name=self.name,
            agency=self.agency,
            description=self.description,
            version=self.version,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            is_partial=self.isPartial,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            codes=[i.to_model(cls) for i in self.hierarchicalCodes],
        )


class JsonHierarchies(Struct, frozen=True):
    """SDMX-JSON payload for hierarchies."""

    codelists: Sequence[JsonCodelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()

    def to_model(self) -> Sequence[Hierarchy]:
        """Returns the requested hierarchy."""
        return [h.to_model(self.codelists) for h in self.hierarchies]


class JsonHierarchyAssociation(MaintainableType, frozen=True):
    """SDMX-JSON payload for a hierarchy association."""

    linkedHierarchy: str = ""
    linkedObject: str = ""
    contextObject: str = ""
    links: Sequence[JsonLink] = ()

    def to_model(
        self,
        hierarchies: Sequence[JsonHierarchy],
        codelists: Sequence[JsonCodelist],
    ) -> HierarchyAssociation:
        """Converts a JsonHierarchyAssocation to a standard association."""
        if hierarchies:
            m = find_by_urn(hierarchies, self.linkedHierarchy).to_model(
                codelists
            )
        else:
            m = self.linkedHierarchy
        lnk = list(
            filter(
                lambda i: hasattr(i, "rel") and i.rel == "UserDefinedOperator",
                self.links,
            )
        )
        return HierarchyAssociation(
            id=self.id,
            name=self.name,
            agency=self.agency,
            hierarchy=m,
            component_ref=self.linkedObject,
            context_ref=self.contextObject,
            description=self.description,
            version=self.version,
            annotations=[a.to_model() for a in self.annotations],
            is_external_reference=self.isExternalReference,
            valid_from=self.validFrom,
            valid_to=self.validTo,
            operator=lnk[0].urn if lnk else None,
        )


class JsonHierarchyMessage(Struct, frozen=True):
    """SDMX-JSON payload for /hierarchy queries."""

    data: JsonHierarchies

    def to_model(self) -> Hierarchy:
        """Returns the requested hierarchy."""
        return self.data.to_model()[0]


class JsonHierarchiesMessage(Struct, frozen=True):
    """SDMX-JSON payload for /hierarchy queries."""

    data: JsonHierarchies

    def to_model(self) -> Sequence[Hierarchy]:
        """Returns the requested hierarchy."""
        return self.data.to_model()


class JsonHierarchyAssociations(Struct, frozen=True):
    """SDMX-JSON payload for hierarchy associations."""

    codelists: Sequence[JsonCodelist] = ()
    hierarchies: Sequence[JsonHierarchy] = ()
    hierarchyAssociations: Sequence[JsonHierarchyAssociation] = ()

    def to_model(self) -> Sequence[HierarchyAssociation]:
        """Returns the requested hierarchy associations."""
        return [
            ha.to_model(self.hierarchies, self.codelists)
            for ha in self.hierarchyAssociations
        ]


class JsonHierarchyAssociationMessage(Struct, frozen=True):
    """SDMX-JSON payload for hierarchy associations messages."""

    data: JsonHierarchyAssociations

    def to_model(self) -> Sequence[HierarchyAssociation]:
        """Returns the requested hierarchy associations."""
        return self.data.to_model()



"""Pandas SDMX Dataset."""

from pysdmx.__extras_check import __check_data_extra
from pysdmx.model.dataset import Dataset

__check_data_extra()

# E402 is needed here to ensure a clear message is used on missing import
import pandas as pd  # noqa: E402


class PandasDataset(Dataset, frozen=False, kw_only=True):
    """A Dataset that is backed by a Pandas DataFrame.

    Args:
        data: Pandas Dataframe to contain SDMX data.
    """

    data: pd.DataFrame



"""Collection of SDMX-JSON schemas for common artefacts."""

from datetime import datetime
from typing import Optional, Sequence, Union

import msgspec

from pysdmx.errors import NotFound
from pysdmx.model import (
    Annotation,
    ArrayBoundaries,
    Codelist,
    Facets,
    Organisation,
)
from pysdmx.model.message import Header
from pysdmx.util import find_by_urn


class JsonLink(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for link objects."""

    href: Optional[str] = None
    urn: Optional[str] = None
    rel: Optional[str] = None
    uri: Optional[str] = None
    type: Optional[str] = None
    title: Optional[str] = None
    hreflang: Optional[str] = None


class JsonAnnotation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for annotations."""

    id: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    value: Optional[str] = None
    text: Optional[str] = None
    links: Sequence[JsonLink] = ()

    def to_model(self) -> Annotation:
        """Converts a JsonAnnotation to a standard Annotation."""
        m = [lnk for lnk in self.links if lnk.rel == "self"]
        lnk = m[0] if len(m) == 1 else None
        if lnk and lnk.href:
            url = lnk.href
        elif lnk and lnk.uri:
            url = lnk.uri
        else:
            url = None
        return Annotation(
            id=self.id,
            title=self.title,
            type=self.type,
            url=url,
            text=self.value if self.value else self.text,
        )


class IdentifiableType(msgspec.Struct, frozen=True):
    """An abstract base type used for all nameable artefacts."""

    id: str
    annotations: Sequence[JsonAnnotation] = ()


class NameableType(msgspec.Struct, frozen=True):
    """An abstract base type used for all nameable artefacts."""

    id: str
    name: str
    description: Optional[str] = None
    annotations: Sequence[JsonAnnotation] = ()


class MaintainableType(
    msgspec.Struct, frozen=True, rename={"agency": "agencyID"}
):
    """An abstract base type used for all maintainable artefacts."""

    agency: str
    id: str
    name: str
    version: str = "1.0"
    isExternalReference: bool = False
    validFrom: Optional[datetime] = None
    validTo: Optional[datetime] = None
    description: Optional[str] = None
    annotations: Sequence[JsonAnnotation] = ()


class ItemSchemeType(MaintainableType, frozen=True):
    """An abstract base type used for all item schemes."""

    isPartial: bool = False


class JsonTextFormat(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for TextFormat."""

    dataType: str
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minValue: Optional[Union[int, float]] = None
    maxValue: Optional[Union[int, float]] = None
    startValue: Optional[Union[int, float]] = None
    endValue: Optional[Union[int, float]] = None
    decimals: Optional[int] = None
    pattern: Optional[str] = None
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    isSequence: bool = False
    isMultilingual: bool = False
    sentinelValues: Optional[Sequence[str]] = None
    timeInterval: Optional[str] = None
    interval: Optional[int] = None


def get_facets(input: JsonTextFormat) -> Facets:
    """Create Facets out of a JsonTextFormat."""
    return Facets(
        min_length=input.minLength,
        max_length=input.maxLength,
        is_sequence=input.isSequence,
        min_value=input.minValue,
        max_value=input.maxValue,
        start_value=input.startValue,
        end_value=input.endValue,
        decimals=input.decimals,
        pattern=input.pattern,
        start_time=input.startTime,
        end_time=input.endTime,
        is_multilingual=input.isMultilingual,
    )


class JsonRepresentation(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for core representation."""

    enumerationFormat: Optional[JsonTextFormat] = None
    enumeration: Optional[str] = None
    format: Optional[JsonTextFormat] = None
    minOccurs: Optional[int] = None
    maxOccurs: Optional[int] = None

    def to_facets(self) -> Optional[Facets]:
        """Return a Facets domain object."""
        fmt = self.enumerationFormat if self.enumeration else self.format
        if fmt and (
            fmt.minLength
            or fmt.maxLength
            or fmt.isSequence
            or fmt.minValue
            or fmt.maxValue
            or fmt.startValue
            or fmt.endValue
            or fmt.decimals
            or fmt.pattern
            or fmt.startTime
            or fmt.endTime
            or fmt.isMultilingual
        ):
            return get_facets(fmt)
        else:
            return None

    def to_enumeration(
        self,
        codelists: Sequence[Codelist],
        valid: Sequence[str],
    ) -> Optional[Codelist]:
        """Returns the list of codes allowed for this component."""
        if self.enumeration:
            try:
                a = find_by_urn(codelists, self.enumeration)
                codes = [c for c in a.codes if not valid or c.id in valid]
                return msgspec.structs.replace(a, items=codes)
            except NotFound:
                # This is OK. In case of schema queries, if a component
                # has both a local and a core representations, only the
                # relevant one (i.e. the local one) will be included in
                # the schema.
                return None
        return None

    def to_array_def(self) -> Optional[ArrayBoundaries]:
        """Returns the array boundaries, if any."""
        if self.maxOccurs and self.maxOccurs > 1:
            m = self.minOccurs if self.minOccurs is not None else 0
            return ArrayBoundaries(m, self.maxOccurs)
        else:
            return None


class JsonHeader(msgspec.Struct, frozen=True):
    """SDMX-JSON payload for message header."""

    id: str
    prepared: datetime
    sender: Organisation
    test: bool = False
    contentLanguages: Sequence[str] = ()
    name: Optional[str] = None
    receivers: Optional[Organisation] = None
    links: Sequence[JsonLink] = ()

    def to_model(self) -> Header:
        """Map to pysdmx header class."""
        return Header(
            id=self.id,
            test=self.test,
            prepared=self.prepared,
            sender=self.sender,
        )



"""Collection of readers for GDS-JSON messages."""

from pysdmx.io.json.gds import messages as msg
from pysdmx.io.serde import GdsDeserializers

deserializers = GdsDeserializers(
    agencies=msg.JsonAgencyMessage,
    catalogs=msg.JsonCatalogMessage,
    sdmx_api=msg.JsonSdmxApiMessage,
    services=msg.JsonServiceMessage,
    urn_resolver=msg.JsonUrnResolverMessage,
)



"""Models for GDS data.

This module defines classes for representing GDS-specific data,
such as agencies, in the SDMX data model.
"""

from typing import List, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.model.__base import MaintainableArtefact
from pysdmx.util import parse_maintainable_urn


class GdsBase(Struct, repr_omit_defaults=True, frozen=True):
    """Base class for all GDS models with a custom __str__ method."""

    def __str__(self) -> str:
        """Custom string representation without the class name."""
        processed_output = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # str is taken as a Sequence, so we need to check it's not a str
            if isinstance(value, Sequence) and not isinstance(value, str):
                # Handle non-empty lists
                if not value:
                    continue
                class_name = value[0].__class__.__name__
                value = f"{len(value)} {class_name.lower()}s"

            processed_output.append(f"{attr}: {value}")
        return f"{', '.join(processed_output)}"

    def __repr__(self) -> str:
        """Custom __repr__ that omits empty sequences."""
        attrs = []
        for attr, value, *_ in self.__rich_repr__():  # type: ignore[misc]
            # Omit empty sequences
            if isinstance(value, (list, tuple, set)) and not value:
                continue
            attrs.append(f"{attr}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(attrs)})"


class GdsEndpoint(GdsBase, frozen=True):
    """Represents a GDS endpoint.

    Attributes:
        api_version: The API version of the endpoint.
        url: The URL of the endpoint.
        comments: Comments about the endpoint.
        message_formats: List of message formats supported by the endpoint.
        rest_resources: List of REST resources available at the endpoint.
    """

    api_version: str
    url: str
    comments: str
    message_formats: List[str]
    rest_resources: List[str]


class GdsServiceReference(GdsBase, frozen=True):
    """Represents a GDS service reference.

    Attributes:
        id: The ID of the service reference.
        name: The name of the service reference.
        urn: The URN of the service reference.
        service: The service associated with the reference.
        description: An optional description of the service reference.
    """

    id: str
    name: str
    urn: str
    service: str
    description: Optional[str] = None

    @property
    def short_urn(self) -> str:
        """Returns a short URN for the ServiceReference."""
        return parse_maintainable_urn(self.service).__str__()


class GdsService(GdsBase, MaintainableArtefact, frozen=True, kw_only=True):
    """Represents a GDS service.

    Attributes:
        agency: The AgencyID of the service's owner.
        base: The base URL of the service.
        endpoints: List of GDS endpoints available at the service.
        authentication: Optional authentication method for the service.
    """

    base: str
    endpoints: List[GdsEndpoint]
    authentication: Optional[str] = None


class GdsCatalog(GdsBase, MaintainableArtefact, frozen=True, kw_only=True):
    """Represents a GDS catalog.

    Attributes:
        agency: The AgencyID of the catalog's owner.
        id: The ID of the catalog.
        urn: The URN of the catalog.
        version: The version of the catalog.
        services: Optional list of GdsService or GdsServiceReference,
        depending on the references parameter, that are associated with the
        catalog.
        endpoints: List of GDS endpoints available at the catalog.
    """

    services: Optional[List[Union[GdsService, GdsServiceReference]]] = None
    endpoints: Optional[List[GdsEndpoint]] = None


class GdsSdmxApi(GdsBase, frozen=True):
    """Represents an SDMX API version.

    Attributes:
        release: The release version of the SDMX API.
        description: A description of the release.
    """

    release: str
    description: str


class ResolverResult(
    GdsBase, frozen=True, rename={"query_response_status_code": "status_code"}
):
    """Represents a single resolver result.

    Attributes:
        api_version: The API version of the resolver result.
        query: The query URL for the resource.
        status_code: The HTTP response code for the query.
    """

    api_version: str
    query: str
    status_code: int


class GdsUrnResolver(GdsBase, frozen=True):
    """Represents the response for a URN resolver query.

    Attributes:
        agency: The agency maintaining the resource.
        resource_id: The ID of the resource.
        version: The version of the resource.
        sdmx_type: The type of SDMX resource.
        resolver_results: A list of resolver results.
    """

    agency: str
    resource_id: str
    version: str
    sdmx_type: str
    resolver_results: List[ResolverResult]



"""Generates the full VTL Transformation Scheme script."""

from typing import Sequence, Union

from pysdmx.__extras_check import __check_vtl_extra
from pysdmx.model import (
    Reference,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperatorScheme,
)


def _process_ruleset_scheme(
    ruleset_schemes: Sequence[Union[RulesetScheme, Reference]],
) -> str:
    """Processes the Ruleset Scheme objects and returns the VTL script."""
    vtl_script = ""
    for ruleset_scheme in ruleset_schemes:
        if isinstance(ruleset_scheme, RulesetScheme):
            for ruleset in ruleset_scheme.items:
                vtl_script += f"{ruleset.ruleset_definition}\n"

    return vtl_script


def _process_udo_scheme(
    udo_schemes: Sequence[Union[UserDefinedOperatorScheme, Reference]],
) -> str:
    """Processes the UDO Scheme objects and returns the VTL script."""
    vtl_script = ""
    for udo_scheme in udo_schemes:
        if isinstance(udo_scheme, UserDefinedOperatorScheme):
            for udo in udo_scheme.items:
                vtl_script += f"{udo.operator_definition}\n"

    return vtl_script


def _process_transformation(transformations: Sequence[Transformation]) -> str:
    """Processes the transformation objects and returns the VTL script."""
    vtl_script = ""
    for transformation in transformations:
        vtl_script += f"{transformation.full_expression}\n"

    return vtl_script


def generate_vtl_script(
    transformation_scheme: TransformationScheme,
    model_validation: bool = False,
    prettyprint: bool = False,
) -> str:
    """Generates the VTL script from a TransformationScheme.

    This method iterates over the TransformationScheme object and its referred
    RulesetSchemes and UserDefinedOperatorSchemes to generate the VTL script
    as string

    The model_validation feature checks if the model object is valid by
    parsing the VTL code inside the definitions.

    The prettyprint feature formats the VTL script in a user-friendly way.

    .. important::
        The prettyprint and model_validation features require
        the pysdmx[vtl] extra.

    Args:
        transformation_scheme: A TransformationScheme object.
        model_validation: A boolean value to check if the model
                        object is valid.
                        if True, the model object is validated.
        prettyprint: A boolean value to check if the generated script
                     is returned formatted.

    Returns:
        A string containing the full VTL Transformation Scheme script.

    """
    if model_validation:
        __check_vtl_extra()
        from pysdmx.toolkit.vtl.validation import model_validations

        model_validations(transformation_scheme)

    vtl_script = ""

    vtl_script += _process_ruleset_scheme(
        transformation_scheme.ruleset_schemes
    )
    vtl_script += _process_udo_scheme(
        transformation_scheme.user_defined_operator_schemes
    )
    vtl_script += _process_transformation(transformation_scheme.items)

    if prettyprint:
        __check_vtl_extra()
        from vtlengine import prettify  # type: ignore[import-untyped]

        return prettify(vtl_script)

    return vtl_script
